"""Utility methods for pdpipe."""

import numpy as np
import pandas as pd


def out_of_place_col_insert(df, series, loc, column_name=None):
    """Returns a new dataframe with given column inserted at given location.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe into which to insert the column.
    series : pandas.Series
        The pandas series to be inserted.
    loc : int
        The location into which to insert the new column.
    column_name : str, default None
        The name to assign the new column. If None, the given series name
        attribute is attempted; if the given series is missing the name
        attribute a ValueError exception will be raised.

    Returns
    -------
    pandas.DataFrame
        The resulting dataframe.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[1, 'a'], [4, 'b']], columns=['a', 'g'])
        >>> ser = pd.Series([7, 5])
        >>> out_of_place_col_insert(df, ser, 1, 'n')
           a  n  g
        0  1  7  a
        1  4  5  b
    """
    if column_name is None:
        if series.name is None:
            raise ValueError("A column name must be supplied if the given "
                             "series is missing the name attribute.")
        column_name = series.name
    inter_df = df.assign(**{column_name: series})
    cols = list(inter_df.columns)
    cols.insert(loc, cols.pop(cols.index(column_name)))
    return inter_df.loc[:, cols]


def get_numeric_column_names(df):
    """Return the names of all columns of numeric type.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to get numeric column names for.

    Returns
    -------
    list of str
        The names of all columns of numeric type.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> data = [[2, 3.2, "acd"], [1, 7.2, "alk"], [8, 12.1, "alk"]]
        >>> df = pd.DataFrame(data, [1,2,3], ["rank", "ph","lbl"])
        >>> sorted(get_numeric_column_names(df))
        ['ph', 'rank']
    """
    num_cols = []
    for colbl, dtype in df.dtypes.to_dict().items():
        if np.issubdtype(dtype, np.number):
            num_cols.append(colbl)
    return num_cols


def per_column_valus_sklearn_transform(df: pd.DataFrame, transform: callable):
    """Applies a 2d-array sklearn transform to 1d values arrays of a dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to transform.
    transform : callable
        The numpy.array-ingesting per-column transform to apply.

    Returns
    -------
    res_df : pandas.DataFrame
        The transformed dataframe.
    """
    return pd.DataFrame(
        data=np.array([
            transform(np.array([series.values]).T)[:, 0]
            for lbl, series in df.iteritems()
        ]).T,
        index=df.index,
        columns=df.columns,
    )



"""Testing performance issue #29.

See: https://github.com/pdpipe/pdpipe/issues/29
"""

import time

import numpy as np
import pandas as pd
import pdpipe as pdp


COLUMNS = [
    "SKC",
    "date",
    "sales",
    "sales_amount",
    "passenger_flow",
    "plus_purchase",
    "activity_level",
    "S",
    "A",
    "B",
    "N",
    "shelf_date",
    "end_of_season",
    "category_group",
    "category",
    "original_price",
    "season",
]


def _original_code():
    start = time.time()
    salesdata = pd.read_csv("processed_salesdata.csv")
    pline = pdp.PdPipeline([
        pdp.Schematize(COLUMNS),
        pdp.ApplyByCols(
            "category_group", lambda x: "tops" if x == "tops" else "other"),
        pdp.ApplyByCols(
            ["date", "shelf_date", "end_of_season"], pd.to_datetime),
        pdp.ApplyToRows(lambda row: pd.Series({
            "standard_amount": row["original_price"] * row["sales"],
            "sales_discount": 0 if (
                row["original_price"] * row["sales"] <= 0
            ) else row["sales_amount"] / (
                (row["original_price"] * row["sales"])
            ),
            "week": int(row["date"].strftime('%W')),
            "days_on_counter": (
                row["date"] - row["shelf_date"]) / np.timedelta64(1, 'D'),
            "life_cycle": (row["end_of_season"] - row["shelf_date"]) / (
                np.timedelta64(1, 'D')),
            "C1": 1 if row["category_group"] == "tops" else 0,
            "C2": 1 if row["category_group"] == "other" else 0,
            "sales": 0 if row["sales"] < 0 else row["sales"],
            "passenger_flow": 0 if row["passenger_flow"] < 0 else (
                row["passenger_flow"]),
            "plus_purchase": 0 if row["plus_purchase"] < 0 else (
                row["plus_purchase"]),
        })),
        pdp.AdHocStage(
            lambda df: df[df["days_on_counter"] <= df["life_cycle"]]),
        pdp.ColDrop("activity_level")
    ])
    salesdata = pline.apply(salesdata, verbose=True, exraise=True)

    salesdata_cumitems = salesdata[
        ["SKC", "date", "sales", "passenger_flow", "plus_purchase"]
    ].sort_values(by=["SKC", "date"]).groupby(['SKC']).cumsum()
    salesdata_cumitems.columns = [
        "total_sales", "total_passenger_flow", "total_plus_purchase"]
    salesdata["total_sales"] = salesdata_cumitems["total_sales"]
    salesdata["total_passenger_flow"] = salesdata_cumitems[
        "total_passenger_flow"]
    salesdata["total_plus_purchase"] = salesdata_cumitems[
        "total_plus_purchase"]
    print("consumed time(s)=", time.time() - start)



"""Wrapper-kind pdpipe pipline stages."""


from pdpipe.core import PdPipelineStage


class FitOnly(PdPipelineStage):
    """A wrapper that applies a stage to input data only when fitting.

    Parameters
    ----------
    stage : PdPipelineStage
        The pipeline stage to operate on input data only when fitting.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[8,'a'],[5,'b']], [1,2], ['num', 'char'])
        >>> stage = pdp.FitOnly(pdp.ColDrop('num'))
        >>> stage(df)
          char
        1    a
        2    b
        >>> df2 = pd.DataFrame([[8,'a'],[5,'b']], [1,2], ['num', 'char'])
        >>> stage(df2)
           num char
        1    8    a
        2    5    b
    """
    _FITONLY_DESC = "Applying, only on fit, the stage: {}"

    def __init__(self, stage, **kwargs):
        self._stage = stage
        desc = FitOnly._FITONLY_DESC.format(stage.description())
        super_kwargs = {
            'desc': desc,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, df):
        if self.is_fitted:
            return True
        return self._stage._prec(df)

    def _fit_transform(self, df, verbose):
        self.is_fitted = True
        return self._stage.fit_transform(df, verbose=verbose)

    def _transform(self, df, verbose):
        if verbose:
            print(
                f"Skipping, because not in fit, "
                f"the stage: {self._stage.description()}"
            )
        return df



"""Column qualifiers for pdpipe.

Most pipeline stages in pdpipe can accept three types of variables as arguments
for the `columns` parameter of their constructor: a single column label, a list
of column labels, or a callable. The first is interpreted as the label of the
single column on which the pipeline stage should operate and the second as a
list of such labels, while the a callable is assumed to determine dynamically
what columns should the stage be applied to. It is thus supplied with the
entire input dataframe, and is expected to return a list of column labels. This
is true for every application of the pipeline stage, both in fit time and in
any future transform.

A naive callable, such as `lambda df: [lbl for lbl in df.columns if lbl[0] ==
'a']`, meant to cause the pipeline stage to operate on any column with a label
starting with the letter 'a', might result in unexpected errors: if after the
pipeline was fitted it gets ― in transform time ― a dataframe with new columns
starting with 'a', it is will transform them as well, which will in turn might
(1) lead to unexpected errors, as the newer columns might not be valid input to
that stage, and (2) lead to a change in schema, which might cause errors down
the pipeline, especially if there's a fitted machine learning model down the
pipeline.

Of course, this might be the desired behaviour ― to tranform coloumns 'alec'
and 'alex' on the first `apply` call and 'alec' and 'apoxy' in transform time ―
but usually this is not the case. In fact, in common machine learning
scenarios ― whether it is fitting pre-processing parameters on the train set
and using the resulting pipeline to transform the test and validation sets, or
fitting the pre-processing pipeline on current data and deploying it to
transform incoming data in production ― we would exepct this criterion to be
applied once, when the pipeline is being fit, and for future calls for it to
only transform the 'alec' and 'alex' columns, ignoring any other columns
starting with 'a' that are newly-encountered in transform time, and also
explicitly fails if either once of the two columns, 'alec' and 'alex', is
missing. This kind of behaviour is the only way to ensure preservation of both
form and semantics of input vectors to our models down the pipeline.

To enable this more sophicticated behaviour this module - `pdpipe.cq` - exposes
a way to easily generate `ColumnQualifier` objects, which are callables that do
exactly what was described above: Apply some criteria to determine a set of
input columns when a pipeline is being fitted, but fixing it afterwards, on
future calls.

Practically, this objects all expose `fit`, `transform` and `fit_transform`
methods, and while the first time they are called the `fit_transform` method is
called, future calls will actually call the `transform` method. Also, since
they already expose this more poweful API, pipeline stages use it to enable
an even more powerful (and quite frankly, expected) behavior: When a pipeline's
`fit_transfrom` or `fit` methods are called, it calls the `fit_transform`
method of the column qualifier it uses, so the qualifier itself is refitted
even if it is already fit. Naturally, if the callable has no `fit_transform`
method the code gracefully backs-off to just applying it, which allows the use
of unfittable functions and lambdas as column qualifiers as well.

Note that any callable can be wrapped in a ColumnQualifier object to achieve
this fittable behaviour. For example, to get a pipeline stage that drops all
columns of data type `numpy.int64`, but also "remembers" that list after fit:

```python
pipeline += pdp.ColDrop(columns=pdp.cq.ColumnQualifier(lambda df: [
  l for l, s in df.iteritems()
  if s.dtype == np.int64
]))
```

ColumnQualifier objects also support the &, ^ and | binary operators -
representing boolean and, xor and or, respectively - and the ~ unary boolean
operator - representing the boolean not operator. Finally, the - binary
operator is implemented to represent the NOT IN non-symetric binary relation
between two qualifiers (the difference operator on the resulting sets).

So for example, to get a qualifier that qualifies all columns that have AT
LEAST two missing values, one can use:

    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame(
    ...    [[None, 1, 2],[None, None, 5]], [1,2], ['ph', 'grade', 'age'])
    >>> cq = ~ pdp.cq.WithAtMostMissingValues(1)
    >>> cq(df)
    ['ph']

While to get a qualifier matching all columns with at most one missing value
AND starting with 'gr' one can use:

    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame(
    ...    [[None, 1, 2],[None, None, 5]], [1,2], ['grep', 'grade', 'age'])
    >>> cq = pdp.cq.WithAtMostMissingValues(1) & pdp.cq.StartWith('gr')
    >>> cq(df)
    ['grade']

And a qualifier that qualifies all columns with no missing values except those
that start with 'b' can be generated with:

    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame(
    ...    [[1, 2, 3, 4],[5, 6, 7, None]], [1,2], ['abe', 'bee', 'cry', 'no'])
    >>> cq = pdp.cq.WithoutMissingValues() - pdp.cq.StartWith('b')
    >>> cq(df)
    ['abe', 'cry']
"""

from .shared import _list_str


class UnfittedColumnQualifierError(Exception):
    """An exception raised when a (non-fit) transform is attempted with an
    unfitted column qualifier.
    """


class ColumnQualifier(object):
    """A fittable qualifier that returns column labels from an input dataframe.

    Parameters
    ----------
    func : callable
        A callable that given an input pandas.DataFrame objects returns a list
        of labels of a subset of the columns of the input dataframe.
    fittable : bool, default True
        If set to false, this qualifier becomes unfittable, and `func` is
        called on every call to transform. True by default.
    subset : bool, default False
        If set to true, fitted qualifiers return the subset of fitted columns
        found in input dataframes during transform, in the order they appeared
        when fitted (NOT in the order they appear in the input dataframe of the
        transform). False by default, which means fitted qualifiers return the
        FULL list of fitted columns, ignoring input dataframes completely on
        transforms. When combined with most pipeline stages, this means the
        stage will fail on its precondition if trying to transform with it a
        dataframe that is missing some values in the fitted qualifier.

    Example
    -------
        >>> import numpy as np; import pdpipe as pdp;
        >>> cq = pdp.cq.ColumnQualifier(lambda df: [
        ...    l for l, s in df.iteritems()
        ...    if s.dtype == np.int64 and l in ['a', 'b', 5]
        ... ])
        >>> cq
        <ColumnQualifier: Qualify columns by function>
        >>> col_drop = pdp.ColDrop(columns=cq)
    """

    def __init__(self, func, fittable=None, subset=None):
        if fittable is None:
            fittable = True
        self._cqfunc = func
        self.__doc__ = func.__doc__
        self._fittable = fittable
        self._subset = subset

    def __call__(self, df):
        """Returns column labels of qualified columns from an input dataframe.

        Parameters
        ----------
        df : pandas.DataFrame
            The input dataframe, from which columns are selected.

        Returns
        -------
        list of objects
            A list of labels of the qualified columns for the input dataframe.
        """
        try:
            return self.transform(df)
        except UnfittedColumnQualifierError:
            return self.fit_transform(df)

    def fit_transform(self, df):
        """Fits this qualifier and returns the labels of the qualifying columns.

        Parameters
        ----------
        df : pandas.DataFrame
            The input dataframe, from which columns are selected.

        Returns
        -------
        list of objects
            A list of labels of the qualified columns for the input dataframe.
        """
        self._columns = self._cqfunc(df)
        return self._columns

    def fit(self, df):
        """Fits this qualifier on the input dataframe.

        Parameters
        ----------
        df : pandas.DataFrame
            The input dataframe, from which columns are selected.

        """
        self.fit_transform(df)

    def transform(self, df):
        """Applies and returns the labels of the qualifying columns.

        Is this ColumnQualifier is fittable, it will return the list of column
        labels that was determined when fitted (or the subset of it that can
        be found in the input datarame), if it's fitted, and throw an exception
        if it is not.

        Parameters
        ----------
        df : pandas.DataFrame
            The input dataframe, from which columns are selected.

        Returns
        -------
        list of objects
            A list of labels of the qualified columns for the input dataframe.
        """
        if not self._fittable:
            return self._cqfunc(df)
        try:
            if self._subset:
                return [x for x in self._columns if x in df.columns]
            return self._columns
        except AttributeError:
            raise UnfittedColumnQualifierError

    def __repr__(self):
        fstr = ''
        if self._cqfunc.__doc__:  # pragma: no cover
            fstr = f' - {self._cqfunc.__doc__}'
        return f"<ColumnQualifier: Qualify columns by function{fstr}>"

    # --- overriding boolean operators ---

    @staticmethod
    def _x_inorderof_y(x, y):
        return [i for i in y if i in x]

    class _AndQualifierFunc(object):
        """A pickle-able AND qualifier class."""

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, df):
            return ColumnQualifier._x_inorderof_y(
                x=set(self.first(df)).intersection(self.second(df)),
                y=df.columns,
            )

    def __and__(self, other):
        try:
            res_func = ColumnQualifier._AndQualifierFunc(
                first=self._cqfunc,
                second=other._cqfunc,
            )
            res_func.__doc__ = (
                f"{self._cqfunc.__doc__ or 'Anonymous qualifier 1'} AND "
                f"{other._cqfunc.__doc__ or 'Anonymous qualifier 2'}"
            )
            return ColumnQualifier(func=res_func)
        except AttributeError:
            return NotImplemented

    class _XorQualifierFunc(object):
        """A pickle-able XOR qualifier class."""

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, df):
            return ColumnQualifier._x_inorderof_y(
                x=set(self.first(df)).symmetric_difference(self.second(df)),
                y=df.columns,
            )

    def __xor__(self, other):
        try:
            res_func = ColumnQualifier._XorQualifierFunc(
                first=self._cqfunc,
                second=other._cqfunc,
            )
            res_func.__doc__ = (
                f"{self._cqfunc.__doc__ or 'Anonymous qualifier 1'} XOR "
                f"{other._cqfunc.__doc__ or 'Anonymous qualifier 2'}"
            )
            return ColumnQualifier(func=res_func)
        except AttributeError:
            return NotImplemented

    class _OrQualifierFunc(object):
        """A pickle-able OR qualifier class."""

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, df):
            return ColumnQualifier._x_inorderof_y(
                x=set(self.first(df)).union(self.second(df)),
                y=df.columns,
            )

    def __or__(self, other):
        try:
            res_func = ColumnQualifier._OrQualifierFunc(
                first=self._cqfunc,
                second=other._cqfunc,
            )
            res_func.__doc__ = (
                f"{self._cqfunc.__doc__ or 'Anonymous qualifier 1'} OR "
                f"{other._cqfunc.__doc__ or 'Anonymous qualifier 2'}"
            )
            return ColumnQualifier(func=res_func)
        except AttributeError:
            return NotImplemented

    class _SubQualifierFunc(object):
        """A pickle-able SUB qualifier class."""

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, df):
            return ColumnQualifier._x_inorderof_y(
                x=set(self.first(df)).difference(self.second(df)),
                y=df.columns,
            )

    def __sub__(self, other):
        try:
            res_func = ColumnQualifier._SubQualifierFunc(
                first=self._cqfunc,
                second=other._cqfunc,
            )
            res_func.__doc__ = (
                f"{self._cqfunc.__doc__ or 'Anonymous qualifier 1'} NOT IN "
                f"{other._cqfunc.__doc__ or 'Anonymous qualifier 2'}"
            )
            return ColumnQualifier(func=res_func)
        except AttributeError:
            return NotImplemented

    class _NotQualifierFunc(object):
        """A pickle-able NOT qualifier class."""

        def __init__(self, cq):
            self.cq = cq

        def __call__(self, df):
            return ColumnQualifier._x_inorderof_y(
                x=set(df.columns).difference(self.cq(df)),
                y=df.columns,
            )

    def __invert__(self):
        res_func = ColumnQualifier._NotQualifierFunc(
            cq=self._cqfunc
        )
        res_func.__doc__ = (
            f"NOT {self._cqfunc.__doc__ or 'Anonymous qualifier'}"
        )
        return ColumnQualifier(func=res_func)


def is_fittable_column_qualifier(obj):
    """Returns True for objects that are fittable ColumnQualifier objects.

    Parameters
    ----------
    obj : object
        The object to examine.

    Returns
    -------
    bool
        True if the given object is an instance of ColumnQualifier and
        fittable, False otherwise.
    """
    return isinstance(obj, ColumnQualifier) and obj._fittable


class AllColumns(ColumnQualifier):
    """Selectes all columns in input dataframes.

    Parameters
    ----------
    **kwargs
        Accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of ColumnQualifier for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[8,1],[5,2]], [1,2], ['a', 'b'])
        >>> cq = pdp.cq.AllColumns()
        >>> cq
        <ColumnQualifier: Qualify all columns>
        >>> cq(df)
        ['a', 'b']
        >>> df2 = pd.DataFrame([[8,1],[5,2]], [1,2], ['b', 'c'])
        >>> cq(df2)
        ['a', 'b']
        >>> cq = pdp.cq.AllColumns(fittable=False)
        >>> cq(df)
        ['a', 'b']
        >>> cq(df2)
        ['b', 'c']
        >>> cq = pdp.cq.AllColumns(subset=True)
        >>> cq(df)
        ['a', 'b']
        >>> cq(df2)
        ['b']
    """

    class _SelectAllColumns(object):

        def __call__(self, df):
            return list(df.columns)

    def __init__(self, **kwargs):
        kwargs['func'] = AllColumns._SelectAllColumns()
        super().__init__(**kwargs)

    def __repr__(self):
        return "<ColumnQualifier: Qualify all columns>"


class ByColumnCondition(ColumnQualifier):
    """A fittable column qualifier based on a per-column condition.

    Parameters
    ----------
    cond : callable
        A callaable that given an input pandas.Series object returns a boolean
        value.
    safe : bool, default False
        If set to True, every call to given condition `cond` is is wrapped in
        a way that interprets every raised exception as a returned False value.
        This is useful when generating qualifiers based on conditions that
        assume a specific datatype for the checked column.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of ColumnQualifier for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame(
        ...    [[1, 2, 'A'],[4, 1, 'C']], [1,2], ['age', 'count', 'grade'])
        >>> cq = pdp.cq.ByColumnCondition(lambda s: s.sum() > 3, safe=True)
        >>> cq(df)
        ['age']
    """

    class _SafeCond(object):

        def __init__(self, cond):
            self.cond = cond

        def __call__(self, series):
            try:
                return self.cond(series)
            except Exception:
                return False

    class _ColumnConditionChecker(object):

        def __init__(self, cond):
            self.cond = cond

        def __call__(self, df):
            return list([
                lbl for lbl, series in df.iteritems()
                if self.cond(series)
            ])

    def __init__(self, cond, safe=False, **kwargs):
        self._cond = cond
        if safe:
            self._cond = ByColumnCondition._SafeCond(cond)
        kwargs['func'] = ByColumnCondition._ColumnConditionChecker(self._cond)
        super().__init__(**kwargs)


class ByLabels(ColumnQualifier):
    """Selectes all columns with the given label or labels.

    Parameters
    ----------
    labels : single label or list-like
        Column labels which qualify.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of ColumnQualifier for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame(
        ...    [[8,'a',5],[5,'b',7]], [1,2], ['num', 'chr', 'nur'])
        >>> cq = pdp.cq.ByLabels('num')
        >>> cq(df)
        ['num']
        >>> cq = pdp.cq.ByLabels(['chr', 'nur'])
        >>> cq(df)
        ['chr', 'nur']
        >>> cq = pdp.cq.ByLabels(['num', 'foo'])
        >>> cq(df)
        ['num']
    """

    class _LabelsQualifierFunc(object):

        def __init__(self, labels):
            self.labels = labels

        def __call__(self, df):
            return [
                lbl for lbl in df.columns
                if lbl in self.labels
            ]

    def __init__(self, labels, **kwargs):
        if isinstance(labels, str) or not hasattr(labels, '__iter__'):
            labels = [labels]
        self._labels = labels
        self._labels_str = _list_str(self._labels)
        cqfunc = ByLabels._LabelsQualifierFunc(self._labels)
        cqfunc.__doc__ = f"Columns with labels in {self._labels_str}"
        self.__doc__ = cqfunc.__doc__
        kwargs['func'] = cqfunc
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<ColumnQualifier: By labels in {self._labels_str}>"


def columns_to_qualifier(columns):
    """Converts the given columns parameter to an equivalent column qualifier.

    Parameters
    ----------
    columns : single label, list-like or callable
        The label, or an iterable of labels, of columns. Alternatively,
        this parameter can be assigned a callable returning an iterable of
        labels from an input pandas.DataFrame. See pdpipe.cq.

    Returns
    -------
    qualifier : ColumnQualifier
        The equivalent ColumnQualifier object.

    Example
    -------
        >>> import pdpipe as pdp;
        >>> pdp.cq.columns_to_qualifier('nu')
        <ColumnQualifier: By labels in nu>
        >>> pdp.cq.columns_to_qualifier(['nu', 'bu'])
        <ColumnQualifier: By labels in nu, bu>
        >>> pdp.cq.columns_to_qualifier(lambda df: [l for l in df.columns])
        <ColumnQualifier: Qualify columns by function>
    """
    if callable(columns):
        if isinstance(columns, ColumnQualifier):
            return columns
        return ColumnQualifier(columns, fittable=False)
    return ByLabels(columns)


class StartWith(ColumnQualifier):
    """Selectes all columns that start with the given string.

    Parameters
    ----------
    prefix : str
        The prefix which qualifies columns.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of ColumnQualifier for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame(
        ...    [[8,'a',5],[5,'b',7]], [1,2], ['num', 'chr', 'nur'])
        >>> cq = pdp.cq.StartWith('nu')
        >>> cq
        <ColumnQualifier: Columns starting with nu>
        >>> cq(df)
        ['num', 'nur']
    """

    @staticmethod
    def _safe_startwith(string, prefix):
        try:
            return string.startswith(prefix)
        except AttributeError:
            return False

    class _StartWithFunc(object):

        def __init__(self, prefix):
            self.prefix = prefix

        def __call__(self, df):
            return [
                lbl for lbl in df.columns
                if StartWith._safe_startwith(lbl, self.prefix)
            ]

    def __init__(self, prefix, **kwargs):
        self._prefix = prefix
        cqfunc = StartWith._StartWithFunc(prefix)
        cqfunc.__doc__ = f"Columns that start with {self._prefix}"
        self.__doc__ = cqfunc.__doc__
        kwargs['func'] = cqfunc
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<ColumnQualifier: Columns starting with {self._prefix}>"


class OfDtypes(ColumnQualifier):
    """Selectes all columns that are of a given dtypes.

    Use `dtypes=np.number` to qualify all numeric columns.

    Parameters
    ----------
    dtypes : object or list of objects
        The dtype or dtypes which qualify columns. Support all valid arguments
        to the `include` parameter of pandas.DataFrame.select_dtypes().
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of ColumnQualifier for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp; import numpy as np;
        >>> df = pd.DataFrame(
        ...    [[8.2,'a',5],[5.1,'b',7]], [1,2], ['ph', 'grade', 'age'])
        >>> cq = pdp.cq.OfDtypes(np.number)
        >>> cq(df)
        ['ph', 'age']
        >>> cq = pdp.cq.OfDtypes([np.number, object])
        >>> cq(df)
        ['ph', 'grade', 'age']
        >>> cq = pdp.cq.OfDtypes(np.int64)
        >>> cq
        <ColumnQualifier: With dtypes in <class 'numpy.int64'>>
        >>> cq(df)
        ['age']
    """

    class _OfDtypeFunc(object):

        def __init__(self, dtypes):
            self.dtypes = dtypes

        def __call__(self, df):
            return list(df.select_dtypes(include=self.dtypes).columns)

    def __init__(self, dtypes, **kwargs):
        self._dtypes = dtypes
        self._dtypes_str = _list_str(self._dtypes)
        cqfunc = OfDtypes._OfDtypeFunc(dtypes)
        cqfunc.__doc__ = f"Columns of dtypes {self._dtypes_str}"
        self.__doc__ = cqfunc.__doc__
        kwargs['func'] = cqfunc
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<ColumnQualifier: With dtypes in {self._dtypes_str}>"


class WithAtMostMissingValues(ColumnQualifier):
    """Selectes all columns with no more than X missing values.

    Parameters
    ----------
    n_missing : int
        The maximum number of missing values with which columns can still
        qualify.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        ColumnQualifier. See the documentation of ColumnQualifier for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp; import numpy as np;
        >>> df = pd.DataFrame(
        ...    [[None, 1, 2],[None, None, 5]], [1,2], ['ph', 'grade', 'age'])
        >>> cq = pdp.cq.WithAtMostMissingValues(1)
        >>> cq
        <ColumnQualifier: With at most 1 missing values>
        >>> cq(df)
        ['grade', 'age']
    """

    class _AtMostFunc(object):

        def __init__(self, n_missing):
            self._n_missing = n_missing

        def __call__(self, df):
            return list(df.columns[df.isna().sum() <= self._n_missing])

    def __init__(self, n_missing, **kwargs):
        self._n_missing = n_missing
        cqfunc = WithAtMostMissingValues._AtMostFunc(n_missing)
        cqfunc.__doc__ = (
            f"Columns with at most {self._n_missing} missing values"
        )
        self.__doc__ = cqfunc.__doc__
        kwargs['func'] = cqfunc
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<ColumnQualifier: " \
               f"With at most {self._n_missing} missing values>"


class WithoutMissingValues(WithAtMostMissingValues):
    """Selectes all columns with no missing values.

    Parameters
    ----------
    **kwargs
        Accepts all keyword arguments of the constructor of ColumnQualifier.
        See the documentation of ColumnQualifier for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp; import numpy as np;
        >>> df = pd.DataFrame(
        ...    [[None, 1, 2],[None, None, 5]], [1,2], ['ph', 'grade', 'age'])
        >>> cq = pdp.cq.WithoutMissingValues()
        >>> cq
        <ColumnQualifier: Without missing values>
        >>> cq(df)
        ['age']
    """

    def __init__(self, **kwargs):
        kwargs['n_missing'] = 0
        super().__init__(**kwargs)

    def __repr__(self):
        return "<ColumnQualifier: Without missing values>"



"""Test nltk pipeline stages."""

import os
import sys

import pytest
import pandas as pd
import pdpipe as pdp


@pytest.mark.first
def test_tokenize():
    df = pd.DataFrame([[3.2, "Kick the baby!"]], [1], ['freq', 'content'])
    tokenize_stage = pdp.TokenizeText('content')
    res_df = tokenize_stage(df)
    assert 'content' in res_df.columns
    assert 'content_tok' not in res_df.columns
    assert res_df['content'][1] == ['Kick', 'the', 'baby', '!']

    tokenize_stage = pdp.TokenizeText('content', drop=False)
    res_df = tokenize_stage(df)
    assert 'content' in res_df.columns
    assert 'content_tok' in res_df.columns
    assert res_df['content'][1] == "Kick the baby!"
    assert res_df['content_tok'][1] == ['Kick', 'the', 'baby', '!']


@pytest.mark.first
def test_untokenize():
    df = pd.DataFrame([[3.2, ['Shake', 'and', 'bake!']]], [1], ['freq', 'txt'])
    untok = pdp.UntokenizeText('txt')
    res_df = untok(df)
    assert 'txt' in res_df.columns
    assert 'txt_untok' not in res_df.columns
    assert res_df['txt'][1] == "Shake and bake!"

    untok = pdp.UntokenizeText('txt', drop=False)
    res_df = untok(df)
    assert 'txt' in res_df.columns
    assert 'txt_untok' in res_df.columns
    assert res_df['txt'][1] == ['Shake', 'and', 'bake!']
    assert res_df['txt_untok'][1] == "Shake and bake!"


@pytest.mark.first
@pytest.mark.skipif(
    (os.name == 'nt') or (sys.platform.startswith('win')),
    reason="nltk has a problem locating resources on windows",
)
def test_remove_stopwords():
    df = pd.DataFrame([[3.2, ['kick', 'the', 'baby']]], [1], ['freq', 'txt'])
    stop = pdp.RemoveStopwords('english', 'txt')
    res_df = stop(df)
    assert 'txt' in res_df.columns
    assert 'txt_nostop' not in res_df.columns
    assert res_df['txt'][1] == ['kick', 'baby']

    stop = pdp.RemoveStopwords('english', 'txt', drop=False)
    res_df = stop(df)
    assert 'txt' in res_df.columns
    assert 'txt_nostop' in res_df.columns
    assert res_df['txt'][1] == ['kick', 'the', 'baby']
    assert res_df['txt_nostop'][1] == ['kick', 'baby']

    stop = pdp.RemoveStopwords(['baby', 'fart'], 'txt', drop=False)
    res_df = stop(df)
    assert 'txt' in res_df.columns
    assert 'txt_nostop' in res_df.columns
    assert res_df['txt'][1] == ['kick', 'the', 'baby']
    assert res_df['txt_nostop'][1] == ['kick', 'the']

    with pytest.raises(TypeError):
        pdp.RemoveStopwords(34, 'txt')



"""Unit test for column qualifiers."""

import pytest
import numpy as np
import pandas as pd
import pdpipe as pdp


NA_DF = pd.DataFrame(
    [[None, 1, 2], [None, None, 5]], [1, 2], ['ph', 'grade', 'age'])

NA_DF2 = pd.DataFrame(
    [[8, 1, 2], [1, 2, 5]], [1, 2], ['ph', 'grade', 'age'])


def test_with_at_most_missing_values():
    cq = pdp.cq.WithAtMostMissingValues(1)
    assert cq(NA_DF) == ['grade', 'age']
    assert cq(NA_DF2) == ['grade', 'age']
    assert cq.fit_transform(NA_DF2) == ['ph', 'grade', 'age']
    # test fit
    assert cq.fit_transform(NA_DF) == ['grade', 'age']
    cq.fit(NA_DF2)
    assert cq(NA_DF) == ['ph', 'grade', 'age']


def test_unfittable_with_at_most_missing_values():
    cq = pdp.cq.WithAtMostMissingValues(1, fittable=False)
    assert cq(NA_DF) == ['grade', 'age']
    assert cq(NA_DF2) == ['ph', 'grade', 'age']


def test_not_operator():
    cq = ~ pdp.cq.WithAtMostMissingValues(1)
    assert cq(NA_DF) == ['ph']


NA_GLBL_DF = pd.DataFrame(
    [[None, 1, 2], [None, None, 5]], [1, 2], ['grep', 'grade', 'age'])


def test_and_operator():
    cq = pdp.cq.WithAtMostMissingValues(1) & pdp.cq.StartWith('gr')
    assert cq(NA_GLBL_DF) == ['grade']


def test_xor_operator():
    cq = pdp.cq.WithAtMostMissingValues(1) ^ pdp.cq.StartWith('gr')
    assert cq(NA_GLBL_DF) == ['grep', 'age']


def test_or_operator():
    cq = pdp.cq.WithAtMostMissingValues(1) | pdp.cq.StartWith('gr')
    assert cq(NA_GLBL_DF) == ['grep', 'grade', 'age']


NA_VARIOUS_FIRST_CHAR_DF = pd.DataFrame(
    [[1, 2, 3, 4], [5, 6, 7, None]], [1, 2], ['abe', 'bee', 'cry', 'no'])


def test_difference_operator():
    cq = pdp.cq.WithoutMissingValues() - pdp.cq.StartWith('b')
    assert cq(NA_VARIOUS_FIRST_CHAR_DF) == ['abe', 'cry']


MIXED_DTYPES_DF = pd.DataFrame(
    [['ab', 2, 1.3], ['bc', 5, 2.2]], [1, 2], ['str', 'int', 'float'])


def test_by_column_condition():
    cq = pdp.cq.ByColumnCondition(lambda x: x.dtype == np.int64)
    assert cq(MIXED_DTYPES_DF) == ['int']


def test_of_dtype():
    print(MIXED_DTYPES_DF)
    print(MIXED_DTYPES_DF.dtypes)
    cq = pdp.cq.OfDtypes(np.number)
    assert cq(MIXED_DTYPES_DF) == ['int', 'float']
    cq = pdp.cq.OfDtypes([np.number, object])
    assert cq(MIXED_DTYPES_DF) == ['str', 'int', 'float']
    cq = pdp.cq.OfDtypes(np.int64)
    assert cq(MIXED_DTYPES_DF) == ['int']


def test_operator_attribute_errors():
    cq = pdp.cq.WithAtMostMissingValues(1)
    with pytest.raises(TypeError):
        cq & 'ad'
    with pytest.raises(TypeError):
        cq | 4
    with pytest.raises(TypeError):
        cq ^ (lambda x: ['a'])
    with pytest.raises(TypeError):
        cq - 32


MIXED_LABELS_DF = pd.DataFrame(
    [[8, 1, 2], [1, 2, 5]], [1, 2], ['ph', 'grad', 48])


def test_start_with():
    cq = pdp.cq.StartWith('g')
    assert cq(MIXED_LABELS_DF) == ['grad']


def test_columns_to_qualifier():
    cq = pdp.cq.StartWith('g')
    b = pdp.cq.columns_to_qualifier(cq)
    assert b == cq



"""Testing MapColVals pipeline stages."""

import datetime

import pandas as pd
import pytest

from pdpipe.col_generation import MapColVals


def _test_df():
    return pd.DataFrame([[1], [3], [2]], ['UK', 'USSR', 'US'], ['Medal'])


def test_mapcolvals():
    """Testing MapColVals pipeline stages."""
    df = _test_df()
    value_map = {1: 'Gold', 2: 'Silver', 3: 'Bronze'}
    res_df = MapColVals('Medal', value_map).apply(df)
    assert res_df['Medal']['UK'] == 'Gold'
    assert res_df['Medal']['USSR'] == 'Bronze'
    assert res_df['Medal']['US'] == 'Silver'


def test_mapcolvals_no_drop():
    """Testing MapColVals pipeline stages."""
    df = _test_df()
    value_map = {1: 'Gold', 2: 'Silver', 3: 'Bronze'}
    res_df = MapColVals('Medal', value_map, drop=False).apply(df)
    assert res_df['Medal']['UK'] == 1
    assert res_df['Medal']['USSR'] == 3
    assert res_df['Medal']['US'] == 2
    assert res_df['Medal_map']['UK'] == 'Gold'
    assert res_df['Medal_map']['USSR'] == 'Bronze'
    assert res_df['Medal_map']['US'] == 'Silver'


def test_mapcolvals_with_res_name():
    """Testing MapColVals pipeline stages."""
    df = _test_df()
    value_map = {1: 'Gold', 2: 'Silver', 3: 'Bronze'}
    res_df = MapColVals('Medal', value_map, result_columns='Metal').apply(df)
    assert res_df['Metal']['UK'] == 'Gold'
    assert res_df['Metal']['USSR'] == 'Bronze'
    assert res_df['Metal']['US'] == 'Silver'


def test_mapcolvals_with_res_name_no_drop():
    """Testing MapColVals pipeline stages."""
    df = _test_df()
    value_map = {1: 'Gold', 2: 'Silver', 3: 'Bronze'}
    map_stage = MapColVals(
        'Medal', value_map, result_columns='Metal', drop=False)
    res_df = map_stage(df)
    assert res_df['Medal']['UK'] == 1
    assert res_df['Medal']['USSR'] == 3
    assert res_df['Medal']['US'] == 2
    assert res_df['Metal']['UK'] == 'Gold'
    assert res_df['Metal']['USSR'] == 'Bronze'
    assert res_df['Metal']['US'] == 'Silver'


def test_mapcolvals_bad_res_name_len():
    """Testing MapColVals pipeline stages."""
    value_map = {1: 'Gold', 2: 'Silver', 3: 'Bronze'}
    with pytest.raises(ValueError):
        map_stage = MapColVals('Medal', value_map, result_columns=['A', 'B'])
        assert isinstance(map_stage, MapColVals)


def _2nd_test_df():
    return pd.DataFrame(
        data=[
            [datetime.timedelta(days=2, minutes=5)],
            [datetime.timedelta(days=3, minutes=10)],
            [datetime.timedelta(days=4, minutes=15)]
        ],
        index=['UK', 'USSR', 'US'],
        columns=['Duration'],
    )


def test_mapcolvals_with_attr_name():
    """Testing MapColVals pipeline stages."""
    df = _2nd_test_df()
    res_df = MapColVals('Duration', 'days').apply(df)
    assert res_df['Duration']['UK'] == 2
    assert res_df['Duration']['USSR'] == 3
    assert res_df['Duration']['US'] == 4


def test_mapcolvals_with_method_name():
    """Testing MapColVals pipeline stages."""
    df = _2nd_test_df()
    res_df = MapColVals('Duration', ('total_seconds', {})).apply(df)
    assert res_df['Duration']['UK'] == df['Duration']['UK'].total_seconds()
    assert res_df['Duration']['USSR'] == df['Duration']['USSR'].total_seconds()
    assert res_df['Duration']['US'] == df['Duration']['US'].total_seconds()



"""Test the DropTokensByLength pipeline stage."""

import pandas as pd
import pdpipe as pdp


DF = pd.DataFrame(
    data=[[4, ["a", "bad", "nice"]], [5, ["good", "university"]]],
    index=[1, 2],
    columns=["age", "text"],
)


def test_drop_tokens_by_len():
    filt = pdp.DropTokensByLength('text', 3, 5)
    res_df = filt(DF)
    assert 'age' in res_df.columns
    assert 'text' in res_df.columns
    assert 'bad' in res_df.loc[1]['text']
    assert 'nice' in res_df.loc[1]['text']
    assert 'a' not in res_df.loc[1]['text']
    assert 'good' in res_df.loc[2]['text']
    assert 'university' not in res_df.loc[2]['text']


def test_drop_tokens_by_len_no_max():
    filt = pdp.DropTokensByLength('text', 3)
    res_df = filt(DF)
    assert 'age' in res_df.columns
    assert 'text' in res_df.columns
    assert 'bad' in res_df.loc[1]['text']
    assert 'nice' in res_df.loc[1]['text']
    assert 'a' not in res_df.loc[1]['text']
    assert 'good' in res_df.loc[2]['text']
    assert 'university' in res_df.loc[2]['text']



"""Testing ValKeep pipeline stages."""

import pandas as pd

from pdpipe.basic_stages import ValKeep


def test_valkeep_with_columns():
    """Testing the ColDrop pipeline stage."""
    df = pd.DataFrame([[1, 4], [4, 5], [5, 11]], [1, 2, 3], ['a', 'b'])
    res_df = ValKeep([4, 5], 'a').apply(df)
    assert 1 not in res_df.index
    assert 2 in res_df.index
    assert 3 in res_df.index


def test_valkeep_with_columns_verbose():
    """Testing the ColDrop pipeline stage."""
    df = pd.DataFrame([[1, 4], [4, 5], [5, 11]], [1, 2, 3], ['a', 'b'])
    res_df = ValKeep([4, 5], 'a').apply(df, verbose=True)
    assert 1 not in res_df.index
    assert 2 in res_df.index
    assert 3 in res_df.index


def test_valkeep_without_columns():
    """Testing the ColDrop pipeline stage."""
    df = pd.DataFrame([[1, 4], [4, 5], [5, 11]], [1, 2, 3], ['a', 'b'])
    res_df = ValKeep([4, 5]).apply(df)
    assert 1 not in res_df.index
    assert 2 in res_df.index
    assert 3 not in res_df.index



"""Unit tests for Condition objects."""

import pytest
# import numpy as np
import pandas as pd
import pdpipe as pdp


NA_DF = pd.DataFrame(
    [[None, 1, 2], [None, None, 5]], [1, 2], ['ph', 'grade', 'age'])

NA_DF2 = pd.DataFrame(
    [[8, 1, 2], [1, 2, 5]], [1, 2], ['ph', 'grade', 'age'])


def test_basic_condition_stuff():
    # check expected behaviour of unfittable conditions
    cond = pdp.cond.HasNoMissingValues()
    assert not cond(NA_DF)
    # fittable should be False by default
    assert cond(NA_DF2)

    # check expected behaviour of fitted conditions
    cond = pdp.cond.HasNoMissingValues(fittable=True)
    cond.fit(NA_DF2)
    assert cond(NA_DF)
    assert not cond.fit_transform(NA_DF)
    assert not cond(NA_DF2)

    # check automatic call to fit_transform on __call__ of unfitted fittable
    cond = pdp.cond.HasNoMissingValues(fittable=True)
    assert cond(NA_DF2)
    assert cond(NA_DF)

    # check correct response for binary operators with unsupported args
    with pytest.raises(TypeError):
        cond & 5
    with pytest.raises(TypeError):
        cond ^ 5
    with pytest.raises(TypeError):
        cond | 5


def test_PerColumnCondition():
    conditions = [
        lambda x: x.isna().sum() > 0,
        lambda x: any([k == 5 for k in x.values]),
    ]
    cond = pdp.cond.PerColumnCondition(
        conditions=conditions,
        conditions_reduce='all',
    )
    assert not cond(NA_DF)

    cond = pdp.cond.PerColumnCondition(
        conditions=conditions,
        conditions_reduce='any',
    )
    print(cond)
    assert cond(NA_DF)

    with pytest.raises(ValueError):
        cond = pdp.cond.PerColumnCondition(
            conditions=conditions,
            conditions_reduce='bad_value',
        )

    with pytest.raises(ValueError):
        cond = pdp.cond.PerColumnCondition(
            conditions=conditions,
            columns_reduce='bad_value',
        )


def test_HasAtMostMissingValues():
    with pytest.raises(ValueError):
        pdp.cond.HasAtMostMissingValues('34')



"""Testing basic pipeline stages."""

import pandas as pd
import pytest

from pdpipe.core import (
    PdPipelineStage,
    FailedPreconditionError,
    FailedPostconditionError,
)


def _test_df():
    return pd.DataFrame(
        data=[[1, 'a'], [2, 'b']],
        index=[1, 2],
        columns=['num', 'char']
    )


class SomeStage(PdPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _prec(self, df):
        return True

    def _transform(self, df, verbose):
        return df


def test_basic_pipeline_stage():
    """Testing the col binner helper class."""
    test_stage = SomeStage()
    assert not test_stage._is_fittable()
    df = _test_df()
    res_df = test_stage.apply(df, verbose=True)
    assert res_df.equals(df)
    res_df = test_stage(df)
    assert res_df.equals(df)
    res_df = test_stage.apply(df, exraise=True)
    assert res_df.equals(df)
    res_df = test_stage.apply(df, exraise=False)
    assert res_df.equals(df)

    # fit_transform
    res_df = test_stage.fit_transform(df)
    assert res_df.equals(df)
    res_df = test_stage.fit_transform(df, exraise=False)
    assert res_df.equals(df)
    res_df = test_stage.fit_transform(df, exraise=False, verbose=False)
    assert res_df.equals(df)
    res_df = test_stage.fit_transform(df, exraise=False, verbose=True)
    assert res_df.equals(df)
    res_df = test_stage.fit_transform(df, exraise=True)
    assert res_df.equals(df)
    res_df = test_stage.fit_transform(df, exraise=True, verbose=False)
    assert res_df.equals(df)
    res_df = test_stage.fit_transform(df, exraise=True, verbose=True)
    assert res_df.equals(df)
    res_df = test_stage.fit_transform(df, verbose=False)
    assert res_df.equals(df)
    res_df = test_stage.fit_transform(df, verbose=True)
    assert res_df.equals(df)

    expected_repr = f"PdPipelineStage: {PdPipelineStage._DEF_DESCRIPTION}"
    assert str(test_stage) == expected_repr
    assert repr(test_stage) == expected_repr


def test_prec_arg():
    test_stage = SomeStage(prec=lambda df: False)
    df = _test_df()
    with pytest.raises(FailedPreconditionError):
        test_stage(df)


class FailStage(PdPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _prec(self, df):
        return False

    def _transform(self, df, verbose):
        return df


def test_fail_pipeline_stage():
    """Testing the col binner helper class."""
    fail_stage = FailStage()
    df = _test_df()
    with pytest.raises(FailedPreconditionError):
        fail_stage.apply(df, verbose=True)
    with pytest.raises(FailedPreconditionError):
        fail_stage.fit_transform(df, verbose=True)
    res_df = fail_stage.fit_transform(df, exraise=False)
    assert res_df.equals(df)


class SilentDropStage(PdPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, colname, **kwargs):
        self.colname = colname
        super().__init__(exraise=False, **kwargs)

    def _prec(self, df):
        return self.colname in df.columns

    def _transform(self, df, verbose):
        return df.drop([self.colname], axis=1)


def test_silent_fail_pipeline_stage():
    """Testing the col binner helper class."""
    silent_fail_stage = SilentDropStage('Tigers')
    df = _test_df()
    res_df = silent_fail_stage.apply(df, verbose=True)
    assert res_df.equals(df)


def test_pipeline_stage_addition_to_int():
    """Testing that """
    silent_fail_stage = SilentDropStage('Tigers')
    with pytest.raises(TypeError):
        silent_fail_stage + 2


def test_skip_arg():
    silent_fail_stage = SilentDropStage('num')
    df = _test_df()
    assert 'num' in df.columns
    res = silent_fail_stage(df)
    assert 'num' not in res.columns

    silent_fail_stage = SilentDropStage('num', skip=lambda df: True)
    res = silent_fail_stage(df)
    assert 'num' in res.columns


def test_stage_name():
    stage = SilentDropStage('Tigers', name='Name')
    assert stage._name == 'Name'
    with pytest.raises(ValueError) as e:
        stage = SilentDropStage('Tigers', name=12345)
    assert str(e.value) == "'name' must be a str, not int."


class FittableDropByCharStage(PdPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, char, **kwargs):
        self.char = char
        super().__init__(**kwargs)

    def _prec(self, df):
        return True

    def _fit_transform(self, df, verbose):
        self.columns = [
            x for x in df.columns
            if x.startswith(self.char)
        ]
        keep_cols = [
            x for x in df.columns
            if x not in self.columns
        ]
        self.is_fitted = True
        return df[keep_cols]

    def _transform(self, df, verbose):
        keep_cols = [
            x for x in df.columns
            if x not in self.columns
        ]
        return df[keep_cols]


def _test_df2():
    return pd.DataFrame(
        data=[[1, 'a'], [2, 'b']],
        index=[1, 2],
        columns=['abo', 'coo']
    )


def _test_df3():
    return pd.DataFrame(
        data=[[1, 'a'], [2, 'b']],
        index=[1, 2],
        columns=['abo', 'aoo']
    )


def test_fittable_stage():
    stage = FittableDropByCharStage('a')
    assert stage._is_fittable()

    res1 = stage(_test_df2())
    assert 'abo' not in res1.columns
    assert 'coo' in res1.columns

    res2 = stage(_test_df3())
    assert 'abo' not in res2.columns
    assert 'aoo' in res2.columns

    stage = FittableDropByCharStage('a')
    res3 = stage(_test_df2())
    assert 'abo' not in res3.columns
    assert 'aoo' not in res3.columns


def _no_a_in_cols(df):
    for lbl in df.columns:
        if 'a' in lbl:
            return False
    return True


def test_fittable_stage_with_postcond():
    stage = FittableDropByCharStage('a', post=_no_a_in_cols)
    assert stage._is_fittable()

    res1 = stage(_test_df2())
    assert 'abo' not in res1.columns
    assert 'coo' in res1.columns

    with pytest.raises(FailedPostconditionError):
        stage(_test_df3())

    with pytest.raises(FailedPostconditionError):
        stage.transform(_test_df3())


def test_failing_postcond():
    stage = SomeStage(post=_no_a_in_cols)

    with pytest.raises(FailedPostconditionError):
        stage(_test_df2())

    res = stage(_test_df2(), exraise=False)
    assert 'abo' in res.columns
    assert 'coo' in res.columns

    with pytest.raises(FailedPostconditionError):
        stage.fit_transform(_test_df2(), exraise=True)

    with pytest.raises(FailedPostconditionError):
        stage.fit(_test_df2(), exraise=True)

    with pytest.raises(FailedPostconditionError):
        stage.transform(_test_df2(), exraise=True)



"""Testing compatability with scikit-learn's Pipelinel objets."""

import pandas as pd
from pdutil.transform import x_y_by_col_lbl
from sklearn.pipeline import make_pipeline
from sklearn.linear_model import LogisticRegression

import pdpipe as pdp


def _train_df():
    return pd.DataFrame(
        data=[
            [3.2, "Pa", "A"],
            [7.2, "Pa", "B"],
            [12.1, "Ko", "B"],
            [18.6, "Pa", "A"],
            [5.3, "Pa", "A"],
            [5.0, "Ko", "A"],
            [17.9, "Ko", "B"],
            [1.3, "Ko", "C"],
        ],
        columns=["ph", "type", "lbl"]
    )


def _test_df():
    return pd.DataFrame(
        data=[
            [4.4, "Ko", "B"],
            [6.1, "Pa", "A"],
            [1.2, "Ko", "B"],
            [1.4, "Ko", "C"],
        ],
        columns=["ph", "lbl", "type"]
    )


def check_sk_pipeline():
    pline = pdp.make_pdpipeline(
        pdp.ApplyByCols("ph", lambda x: x - 1),
        # pdp.Bin({"ph": [0, 3, 5, 12]}),
        pdp.Encode(["type", "lbl"]),
    )
    print(pline)

    model_pline = make_pipeline(
        pdp.FreqDrop(2, "lbl"),
        LogisticRegression(),
    )
    print(model_pline)

    train = _train_df()
    res_train = pline(train)
    print(f"Processed train set: {res_train}")
    x_train, y_train = x_y_by_col_lbl(res_train, "lbl")
    model_pline = model_pline.fit(x_train, y_train)
    print(f"Fitted model pipeline: {model_pline}")

    test = _test_df()
    res_test = pline(test)
    print(f"Processed test set: {res_test}")
    x_test, y_test = x_y_by_col_lbl(res_test, "lbl")
    predictions = model_pline.predict(x_test)
    print(f"predictions: {predictions}")


if __name__ == "__main__":
    check_sk_pipeline()



"""Testing basic pipeline stages."""

import math

import pandas as pd

from pdpipe.col_generation import Bin


def test_col_binner():
    """Testing the col binner helper class."""
    binner = Bin._get_col_binner([0, 5])
    assert binner(-math.inf) == '<0'
    assert binner(-4) == '<0'
    assert binner(0) == '0-5'
    assert binner(1) == '0-5'
    assert binner(4.99) == '0-5'
    assert binner(5) == '5≤'
    assert binner(232) == '5≤'
    assert binner(math.inf) == '5≤'


def test_bin_verbose():
    """Basic binning test."""
    df = pd.DataFrame([[-3], [4], [5], [9]], [1, 2, 3, 4], ['speed'])
    bin_stage = Bin({'speed': [5]}, drop=False)
    res_df = bin_stage.apply(df, verbose=True)
    assert 'speed_bin' in res_df.columns
    assert res_df['speed_bin'][1] == '<5'
    assert res_df['speed_bin'][2] == '<5'
    assert res_df['speed_bin'][3] == '5≤'
    assert res_df['speed_bin'][4] == '5≤'


def test_bin_drop():
    """Basic binning test."""
    df = pd.DataFrame([[-3], [4], [5], [9]], [1, 2, 3, 4], ['speed'])
    bin_stage = Bin({'speed': [5]}, drop=True)
    res_df = bin_stage.apply(df, verbose=True)
    assert 'speed_bin' not in res_df.columns
    assert res_df['speed'][1] == '<5'
    assert res_df['speed'][2] == '<5'
    assert res_df['speed'][3] == '5≤'
    assert res_df['speed'][4] == '5≤'


def test_bin_two_col():
    """Basic binning test."""
    df = pd.DataFrame([[-3, 9], [4, 2], [5, 1], [9, 5]], columns=['s', 'p'])
    bin_stage = Bin({'s': [5], 'p': [5]}, drop=False)
    res_df = bin_stage.apply(df, verbose=True)
    assert 's_bin' in res_df.columns
    assert res_df['s_bin'][0] == '<5'
    assert res_df['s_bin'][1] == '<5'
    assert res_df['s_bin'][2] == '5≤'
    assert res_df['s_bin'][3] == '5≤'
    assert 'p_bin' in res_df.columns
    assert res_df['p_bin'][0] == '5≤'
    assert res_df['p_bin'][1] == '<5'
    assert res_df['p_bin'][2] == '<5'
    assert res_df['p_bin'][3] == '5≤'



"""Testing basic pipeline stages."""

import pytest
import pandas as pd
# from numpy.testing import assert_approx_equal

from pdpipe.sklearn_stages import Scale
from pdpipe.exceptions import PipelineApplicationError


def _some_df1():
    return pd.DataFrame(
        data=[[3.2, 0.3, "A"], [7.2, 0.35, "B"], [12.1, 0.29, "C"]],
        index=[1, 2, 3],
        columns=["ph", "gt", "lbl"],
    )


def _some_df1b():
    return pd.DataFrame(
        data=[[3.8, 0.45, "A"], [7.7, 0.31, "B"], [11.15, 0.33, "C"]],
        index=[1, 2, 3],
        columns=["ph", "gt", "lbl"],
    )


def _bad_dtype_df1():
    return pd.DataFrame(
        data=[[3.2, "H", "A"], [7.2, "I", "B"], [12.1, "P", "C"]],
        index=[1, 2, 3],
        columns=["ph", "gt", "lbl"],
    )


def _some_df2():
    return pd.DataFrame(
        data=[[3.2, 0.3], [7.2, 0.35], [12.1, 0.29]],
        index=[1, 2, 3],
        columns=["ph", "gt"],
    )


def _some_df2b():
    return pd.DataFrame(
        data=[[3.8, 0.45], [7.7, 0.31], [11.15, 0.33]],
        index=[1, 2, 3],
        columns=["ph", "gt"],
    )


def test_scale():
    df = _some_df2()
    scale_stage = Scale("StandardScaler")
    res_df = scale_stage(df)
    assert "ph" in res_df.columns
    assert "gt" in res_df.columns
    assert res_df["ph"][1] < df["ph"][1]

    # see only transform (no fit) when already fitted
    df2 = _some_df2b()
    res_df2 = scale_stage(df2)
    assert "ph" in res_df2.columns
    assert "gt" in res_df2.columns
    assert res_df2["ph"][1] < df2["ph"][1]
    assert res_df["ph"][1] < res_df2["ph"][1]

    # check fit_transform when already fitted
    df3 = _some_df2b()
    res_df3 = scale_stage.fit_transform(df2)
    assert "ph" in res_df3.columns
    assert "gt" in res_df3.columns
    assert res_df3["ph"][1] < df3["ph"][1]
    assert res_df3["ph"][1] < res_df2["ph"][1]


def test_scale_with_exclude_cols():
    df = _some_df1()
    scale_stage = Scale("StandardScaler", exclude_columns=["lbl"])
    res_df = scale_stage(df)
    assert list(res_df.columns) == ["ph", "gt", "lbl"]
    assert "ph" in res_df.columns
    assert "gt" in res_df.columns
    assert res_df["ph"][1] < df["ph"][1]

    # see only transform (no fit) when already fitted
    df2 = _some_df1b()
    res_df2 = scale_stage(df2)
    assert "ph" in res_df2.columns
    assert "gt" in res_df2.columns
    assert df['ph'][1] < df2['ph'][1]
    assert res_df2["ph"][1] < df2["ph"][1]
    assert res_df["ph"][1] < res_df2["ph"][1]

    # check fit_transform when already fitted
    df3 = _some_df1b()
    res_df3 = scale_stage.fit_transform(df2)
    assert "ph" in res_df3.columns
    assert "gt" in res_df3.columns
    assert res_df3["ph"][1] < df3["ph"][1]
    assert res_df3["ph"][1] < res_df2["ph"][1]


def test_scale_with_exclude():
    """Basic binning test."""
    df = _some_df2()
    scale_stage = Scale("StandardScaler", with_std=False)
    res_df = scale_stage(df)
    assert "ph" in res_df.columns
    assert "gt" in res_df.columns


def test_scale_fit_transform_exception():
    df1 = _some_df1()
    scale_stage = Scale("StandardScaler", columns=['ph', 'lbl'])
    with pytest.raises(PipelineApplicationError):
        scale_stage(df1)


def test_scale_transform_exception():
    df1 = _some_df1()
    scale_stage = Scale("StandardScaler", exmsg="ERR")
    scale_stage(df1)

    # test transform exception
    df2 = _bad_dtype_df1()
    with pytest.raises(PipelineApplicationError):
        scale_stage(df2)


def _some_df3():
    return pd.DataFrame(
        data=[[1, 1], [2, 4]],
        index=[1, 2],
        columns=["a", "b"],
    )


def _some_df3b():
    return pd.DataFrame(
        data=[[1, 1], [2, 3]],
        index=[1, 2],
        columns=["a", "b"],
    )


def test_scale_with_joint():
    df = _some_df3()
    scale_stage = Scale("MinMaxScaler")
    res_df = scale_stage(df)
    assert "a" in res_df.columns
    assert "b" in res_df.columns
    assert res_df["a"][2] == 1
    assert res_df["b"][2] == 1

    df = _some_df3()
    scale_stage = Scale("MinMaxScaler", joint=True)
    res_df = scale_stage(df)
    assert "a" in res_df.columns
    assert "b" in res_df.columns
    assert res_df["a"][2] == (1 / 3)
    assert res_df["b"][2] == 1

    df = _some_df3b()
    res_df = scale_stage(df)
    assert "a" in res_df.columns
    assert "b" in res_df.columns
    assert (res_df >= 1).sum().sum() == 0



"""Testing shared functions of pdpipe."""

from pdpipe.shared import (
    _interpret_columns_param,
    _list_str,
    _get_args_list,
    _identity_function,
)


def test_interpret_columns_param():
    assert _interpret_columns_param('a') == ['a']
    assert _interpret_columns_param(5) == [5]
    assert _interpret_columns_param([1, 2]) == [1, 2]


def test_list_str():
    assert _list_str(None) is None
    assert _list_str(['a', 'b']) == 'a, b'
    assert _list_str('a') == 'a'
    assert _list_str((1, 2)) == '1, 2'
    assert _list_str(5) == 5


def foo(a, b):
    return a + b


def test_get_args_list():
    print(_get_args_list(foo))
    assert _get_args_list(foo) == ['a', 'b']


def test_identity_function():
    assert _identity_function(5) == 5



"""Defines pipelines for processing Pandas.DataFrame-based datasets.

>>> import pdpipe as pdp
>>> pipeline = pdp.ColDrop('Name') + pdp.Bin({'Speed': [0,5]})
>>> pipeline = pdp.ColDrop('Name').Bin({'Speed': [0,5]}, drop=True)

## Creating pipeline stages that operate on column subsets

Many pipeline stages in pdpipe operate on a subset of columns, allowing the
caller to deteremine this subset by either providing a fixed set of column
labels or by providing a callable that determines the column subset dynamically
from input dataframes. The `pdpipe.cq` module addresses a unique but important
use case of fittable column qualifier, which dynamically extract a column
subset on stage fit time, but keep it fixed for future transformations.

As a general rule, every pipeline stage in pdpipe that supports the `columns`
parameter should inherently support fittable column qualifier, and generally
the correct interpretation of both single and multiple labels as arguments. To
unify the implementation of such functionality, and ease of creation of new
pipeline stages, such columns shoul be created by extending the
ColumnsBasedPipelineStage base class, found in this module (`pdpipe.core`).

The main interface of sub-classes of this base class with it is through the
`columns`, `exclude_columns` and `none_columns` constructor arguments, and the
"private" `_get_columns(df, fit)` method:

    * Any extending subclass should accept the `columns` constructor parameter
      and forward it, without transforming it, to the constructor of
      ColumnsBasedPipelineStage. E.g.
      `super().__init__(columns=columns, **kwargs)`. See the implementation of
      any such extending class for a more complete example.

    * Extending subclasses can decide if they want to expose the
      `exclude_columns` parameter or not. Note that most of its functionality
      can anyway be gained by providing the `columns` parameter with a column
      qualifier object that is a difference between two column qualifiers; e.g.
      `columns=cq.OfDtype(np.number) - cq.OfDtype(np.int64)` is equivalent to
      providing `columns=cq.OfDtype(np.number),
      exclude_columns=cq.OfDtype(np.int64)`. However, exposing the
      `exclude_columns` parameter can allow for specific unique behaviours; for
      example, if the `none_columns` parametet - which configures the behavior
      when `columns` is provided with `None` - is set with
      a `cq.OfDtypes('category')` column qualifier, which means that all
      categorical columns are selected when `columns=None`, then exposing
      `exclude_columns` allows easy specification of the "all categorical
      columns except X" by just giving a column qualifier capturing X to
      `exclude_columns`, instead of having to reconstruct the default column
      qualifier by hand and substract from it the one representing X.

    * When wishing to get the subset of columns to operate on, in
      `fit_transform` or `transform` time, it is attained by calling
      `self._get_columns(df, fit=True)` (or with `fit=False` if just
      transforming), providing it the input dataframe.

    * Additionally, to get a description and application message with a nice
      string representation of the list of columns to operate on, the
      `desc_temp` constructor parameter of ColumnsBasedPipelineStage can be
      provided with a format string with a place holder where the column list
      should go. E.g. `"Drop columns {}"` for the DropCol pipeline stage.

There are two correct ways to extend it, depending on whether the pipeline
stage you're creating is inherently fittable or not:

    1. If the stage is NOT inherently fittable, then the ability to accept
       fittable column qualifier objects makes it so. However, to enable
       extending subclasses to implement their transformation using a single
       method, they can simply implement the abstract method
       `_transformation(self, df, verbose, fit)`. It should treat the `df` and
       `verbose` parameters normally, but forward the `fit` parameter to the
       `_get_columns` method when calling it. This is enough to get a pipeline
       stage with the desired behavior, with the super-class handling all the
       fit/transform functionality.

    2. If the stage IS inherently fittable, then do not use the
       `_transformation` abstract method (it has to be implemented, so just
       have it raise a NotImplementedError). Instead, simply override the
       `_fit_transform` and `_transform` method of ColumnsBasedPipelineStage,
       calling the `fit` parameter of the `_get_columns` method with the
       correct arguement: `True` when fit-transforming and `False` when
       transforming.

Again, taking a look at the VERY concise implementation of simple columns-based
stages, like ColDrop or ValDrop, will probably make things clearer, and you can
use those implementations as a template for yours.
"""

import sys
import abc
import time
import inspect
import collections
import textwrap

try:
    from pympler.asizeof import asizeof
except ImportError:
    from sys import getsizeof as asizeof

from .cq import is_fittable_column_qualifier, AllColumns
from .shared import _get_args_list
from .exceptions import (
    FailedPreconditionError,
    FailedPostconditionError,
    UnfittedPipelineStageError,
    PipelineApplicationError
)


# === loading stage attributes ===

def __get_append_stage_attr_doc(class_obj):
    doc = class_obj.__doc__
    first_line = doc[0:doc.find('.') + 1]
    if "An" in first_line:
        new_first_line = first_line.replace("An", "Creates and adds an", 1)
    else:
        new_first_line = first_line.replace("A", "Creates and adds a", 1)
    new_first_line = new_first_line[0:-1] + (
        " to this pipeline stage.")
    return doc.replace(first_line, new_first_line, 1)


def __load_stage_attribute__(class_obj):

    def _append_stage_func(self, *args, **kwds):
        # self is always a PdPipelineStage
        return self + class_obj(*args, **kwds)
    _append_stage_func.__doc__ = __get_append_stage_attr_doc(class_obj)
    _append_stage_func.__name__ = class_obj.__name__  # .lower()
    _append_stage_func.__signature__ = inspect.signature(class_obj.__init__)
    setattr(PdPipelineStage, class_obj.__name__, _append_stage_func)

    # unbound_method = types.MethodType(_append_stage_func, class_obj)
    # setattr(class_obj, class_obj.__name__, unbound_method)


def __load_stage_attributes_from_module__(module_name):
    module_obj = sys.modules[module_name]
    for name, obj in inspect.getmembers(module_obj):
        if inspect.isclass(obj) and obj.__module__ == module_name:
            class_obj = getattr(module_obj, name)
            if issubclass(class_obj, PdPipelineStage) and (
                    class_obj.__name__ != 'PdPipelineStage'):
                __load_stage_attribute__(class_obj)


# === basic classes ===

class PdpApplicationContext(dict):
    """An object encapsulating the application context of a pipeline.

    It is meant to communicate data, information and variables between
    different stages of a pipeline.

    Parameters
    ----------
    fit_context : PdpApplicationContext, optional
        Another application context object, representing the application
        context of a previous fit of the pipelline this application context
        is initialized for. Optional.
    """

    def __init__(self, fit_context=None):
        self.__locked__ = False
        self._fit_context__ = fit_context

    def __setitem__(self, key, value):
        if not self.__locked__:
            super().__setitem__(key, value)

    def __delitem__(self, key):
        if not self.__locked__:
            super().__delitem__(key)

    def pop(self, key, default):
        """If key is in the dictionary, remove it and return its value, else
        return default. If default is not given and key is not in the
        dictionary, a KeyError is raised.
        """
        if not self.__locked__:
            return super().pop(key, default)
        return super().__getitem__(key)

    def clear(self):
        """Remove all items from the dictionary."""
        if not self.__locked__:
            super().clear()

    def popitem(self):
        """Not implemented!"""
        raise NotImplementedError

    def update(self, other):
        """Update the dictionary with the key/value pairs from other,
        overwriting existing keys. Return None.
        update() accepts either another dictionary object or an iterable of
        key/value pairs (as tuples or other iterables of length two). If
        keyword arguments are specified, the dictionary is then updated with
        those key/value pairs: d.update(red=1, blue=2).
        """
        if not self.__locked__:
            super().update(other)

    def lock(self):
        """Locks this application context for changes."""
        self.__locked__ = True

    def fit_context(self):
        """Returns a locked PdpApplicationContext object of a previous fit."""
        return self._fit_context__


class PdPipelineStage(abc.ABC):
    """A stage of a pandas DataFrame-processing pipeline.

    Parameters
    ----------
    exraise : bool, default True
        If true, a pdpipe.FailedPreconditionError is raised when this
        stage is applied to a dataframe for which the precondition does
        not hold. Otherwise the stage is skipped. Additionally, if true, a
        pdpipe.FailedPostconditionError is raised if an expected post-codnition
        does not hold for an output dataframe (after pipeline application).
        Otherwise pipeline application continues uninterrupted.
    exmsg : str, default None
        The message of the exception that is raised on a failed
        precondition if exraise is set to True. A default message is used
        if None is given.
    desc : str, default None
        A short description of this stage, used as its string representation.
        A default description is used if None is given.
    prec : callable, default None
        This can be assigned a callable that returns boolean values for input
        dataframes, which will be used to determine whether input dataframes
        satisfy the preconditions for this pipeline stage (see the `exraise`
        parameter for the behaviour of failed preconditions). See pdp.cond for
        more information on specialised Condition objects.
    post : callable, default None
        This can be assigned a callable that returns boolean values for input
        dataframes, which will be used to determine whether input dataframes
        satisfy the postconditions for this pipeline stage (see the `exraise`
        parameter for the behaviour of failed postconditions). See pdp.cond for
        more information on specialised Condition objects.
    skip : callable, default None
        This can be assigned a callable that returns boolean values for input
        dataframes, which will be used to determine whether this stage should
        be skipped for input dataframes. See pdp.cond for more information on
        specialised Condition objects.
    name : str, default ''
        The name of this stage. Pipelines can be sliced by this name.

    Attributes
    ----------
    fit_context : PdpApplicationContext
        An application context object that is only re-initialized before
        `fit_transform` calls, and is locked after pipeline application. It is
        injected into the PipelineStage by the encapsulating pipeline object.
    application_context : PdpApplicationContext
        An application context object that is re-initialized before every
        pipeline application (so, also during transform operations of fitted
        pipelines), and is locked after pipeline application.It is injected
        into the PipelineStage by the encapsulating pipeline object.
    """

    _DEF_EXC_MSG = 'Precondition failed in stage {}!'
    _DEF_DESCRIPTION = 'A pipeline stage.'
    _INIT_KWARGS = ['exraise', 'exmsg', 'desc', 'prec', 'skip', 'name']

    def __init__(self, exraise=True, exmsg=None, desc=None, prec=None,
                 post=None, skip=None, name=''):
        if not isinstance(name, str):
            raise ValueError(
                f"'name' must be a str, not {type(name).__name__}."
            )
        if desc is None:
            desc = PdPipelineStage._DEF_DESCRIPTION
        if exmsg is None:
            exmsg = PdPipelineStage._DEF_EXC_MSG.format(desc)

        self._exraise = exraise
        self._exmsg = exmsg
        self._exmsg_post = exmsg.replace(
            'precondition', 'postcondition').replace(
            'Precondition', 'Postcondition')
        self._desc = desc
        self._prec_arg = prec
        self._post_arg = post
        self._skip = skip
        self._appmsg = f"{name + ': ' if name else ''}{desc}"
        self._name = name
        self.fit_context: PdpApplicationContext = None
        self.application_context: PdpApplicationContext = None
        self.is_fitted = False

    @classmethod
    def _init_kwargs(cls):
        return cls._INIT_KWARGS

    @abc.abstractmethod
    def _prec(self, df):  # pylint: disable=R0201,W0613
        """Returns True if this stage can be applied to the given dataframe."""
        raise NotImplementedError

    def _compound_prec(self, df):
        if self._prec_arg:
            return self._prec_arg(df)
        return self._prec(df)

    def _post(self, df):  # pylint: disable=R0201,W0613
        """Returns True if this stage resulted in an expected output frame."""
        return True

    def _compound_post(self, df):
        if self._post_arg:
            return self._post_arg(df)
        return self._post(df)

    def _fit_transform(self, df, verbose):
        """Fits this stage and transforms the input dataframe."""
        return self._transform(df, verbose)

    def _is_fittable(self):
        if self.__class__._fit_transform == PdPipelineStage._fit_transform:
            return False
        return True

    @abc.abstractmethod
    def _transform(self, df, verbose):
        """Transforms the given dataframe without fitting this stage."""
        raise NotImplementedError("_transform method not implemented!")

    def apply(self, df, exraise=None, verbose=False):
        """Applies this pipeline stage to the given dataframe.

        If the stage is not fitted fit_transform is called. Otherwise,
        transform is called.

        Parameters
        ----------
        df : pandas.DataFrame
            The dataframe to which this pipeline stage will be applied.
        exraise : bool, default None
            Override preconditions and postconditions behaviour for this call.
            If None, the default behaviour of this stage is used, as determined
            by the exraise constructor parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            is checked but before the application of the pipeline stage.
            Defaults to False.

        Returns
        -------
        pandas.DataFrame
            The resulting dataframe.
        """
        if exraise is None:
            exraise = self._exraise
        if self._skip and self._skip(df):
            return df
        if self._compound_prec(df=df):
            if verbose:
                msg = '- ' + '\n  '.join(textwrap.wrap(self._appmsg))
                print(msg, flush=True)
            if self.is_fitted:
                res_df = self._transform(df, verbose=verbose)
            else:
                res_df = self._fit_transform(df, verbose=verbose)
            if exraise and not self._compound_post(df=res_df):
                raise FailedPostconditionError(self._exmsg_post)
            return res_df
        if exraise:
            raise FailedPreconditionError(self._exmsg)
        return df

    __call__ = apply

    def fit_transform(self, X, y=None, exraise=None, verbose=False):
        """Fits this stage and transforms the given dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to transform and fit this pipeline stage by.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Override preconditions and postconditions behaviour for this call.
            If None, the default behaviour of this stage is used, as determined
            by the exraise constructor parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            is checked but before the application of the pipeline stage.
            Defaults to False.

        Returns
        -------
        pandas.DataFrame
            The resulting dataframe.
        """
        if exraise is None:
            exraise = self._exraise
        if self._compound_prec(X):
            if verbose:
                msg = '- ' + '\n  '.join(textwrap.wrap(self._appmsg))
                print(msg, flush=True)
            res_df = self._fit_transform(X, verbose=verbose)
            if exraise and not self._compound_post(df=res_df):
                raise FailedPostconditionError(self._exmsg_post)
            return res_df
        if exraise:
            raise FailedPreconditionError(self._exmsg)
        return X

    def fit(self, X, y=None, exraise=None, verbose=False):
        """Fits this stage without transforming the given dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to be transformed.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Override preconditions and postconditions behaviour for this call.
            If None, the default behaviour of this stage is used, as determined
            by the exraise constructor parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            is checked but before the application of the pipeline stage.
            Defaults to False.

        Returns
        -------
        pandas.DataFrame
            The resulting dataframe.
        """
        if exraise is None:
            exraise = self._exraise
        if self._compound_prec(X):
            if verbose:
                msg = '- ' + '\n  '.join(textwrap.wrap(self._appmsg))
                print(msg, flush=True)
            res_df = self._fit_transform(X, verbose=verbose)
            if exraise and not self._compound_post(df=res_df):
                raise FailedPostconditionError(self._exmsg_post)
            return X
        if exraise:
            raise FailedPreconditionError(self._exmsg)
        return X

    def transform(self, X, y=None, exraise=None, verbose=False):
        """Transforms the given dataframe without fitting this stage.

        If this stage is fittable but is not fitter, an
        UnfittedPipelineStageError is raised.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to be transformed.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Override preconditions and postconditions behaviour for this call.
            If None, the default behaviour of this stage is used, as determined
            by the exraise constructor parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            is checked but before the application of the pipeline stage.
            Defaults to False.

        Returns
        -------
        pandas.DataFrame
            The resulting dataframe.
        """
        if exraise is None:
            exraise = self._exraise
        if self._prec(X):
            if verbose:
                msg = '- ' + '\n  '.join(textwrap.wrap(self._appmsg))
                print(msg, flush=True)
            if self._is_fittable():
                if self.is_fitted:
                    res_df = self._transform(X, verbose=verbose)
                    if exraise and not self._compound_post(df=res_df):
                        raise FailedPostconditionError(self._exmsg_post)
                    return res_df
                raise UnfittedPipelineStageError(
                    "transform of an unfitted pipeline stage was called!")
            res_df = self._transform(X, verbose=verbose)
            if exraise and not self._compound_post(df=res_df):
                raise FailedPostconditionError(self._exmsg_post)
            return res_df
        if exraise:
            raise FailedPreconditionError(self._exmsg)
        return X

    def __add__(self, other):
        if isinstance(other, PdPipeline):
            return PdPipeline([self, *other._stages])
        if isinstance(other, PdPipelineStage):
            return PdPipeline([self, other])
        return NotImplemented

    def __str__(self):
        return f"PdPipelineStage: {self._desc}"

    def __repr__(self):
        return self.__str__()

    def description(self):
        """Returns the description of this pipeline stage"""
        return self._desc

    def _mem_str(self):
        total = asizeof(self)
        lines = []
        for a in dir(self):
            if not a.startswith('__'):
                att = getattr(self, a)
                if not callable(att):
                    size = asizeof(att)
                    if size > 500000:  # pragma: no cover
                        lines.append('  - {}, {:.2f}Mb ({:0>5.2f}%)\n'.format(
                            a, size / 1000000, 100 * size / total))
                    elif size > 1000:  # pragma: no cover
                        lines.append('  - {}, {:.2f}Kb ({:0>5.2f}%)\n'.format(
                            a, size / 1000, 100 * size / total))
                    else:
                        lines.append('  - {}, {}b ({:0>5.2f}%)\n'.format(
                            a, size, 100 * size / total))
        return ''.join(lines)


class ColumnsBasedPipelineStage(PdPipelineStage):
    """A pipeline stage that operates on a subset of dataframe columns.

    Parameters
    ---------
    columns : object, iterable or callable
        The label, or an iterable of labels, of columns to use. Alternatively,
        this parameter can be assigned a callable returning an iterable of
        labels from an input pandas.DataFrame. See pdpipe.cq.
    exclude_columns : object, iterable or callable, optional
        The label, or an iterable of labels, of columns to exclude, given the
        `columns` parameter. Alternatively, this parameter can be assigned a
        callable returning a labels iterable from an input pandas.DataFrame.
        See pdpipe.cq. Optional. By default no columns are excluded.
    desc_temp : str, optional
        If given, assumed to be a format string, and every appearance of {} in
        it is replaced with an appropriate string representation of the columns
        parameter, and is used as the pipeline description. Ignored if `desc`
        is provided.
    none_columns : iterable, callable or str, default 'error'
        Determines how None values supplied to the 'columns' parameter should
        be handled. If set to 'error', the default, a ValueError is raised if
        None is encountered. If set to 'all', it is interpreted to mean all
        columns of input dataframes should be operated on. If an iterable is
        provided it is interpreted as the default list of columns to operate on
        when `columns=None`. If a callable is provided, it is interpreted as
        the default column qualifier that determines input columns when
        `columns=None`.
    **kwargs
        Additionally supports all constructor parameters of PdPipelineStage.
    """

    @staticmethod
    def _interpret_columns_param(columns, none_error=False, none_columns=None):
        """Interprets the value provided to the columns parameter and returns
        a list version of it - if needed - a string representation of it.
        """
        if columns is None:
            if none_error:
                raise ValueError((
                    'None is not a valid argument for the columns parameter of'
                    ' this pipeline stage.'))
            return ColumnsBasedPipelineStage._interpret_columns_param(
                columns=none_columns)
        if isinstance(columns, str):
            # always check str first, because it has __iter__
            return [columns], columns
        if callable(columns):
            # if isinstance(columns, ColumnQualifier):
            #     return columns, columns.__repr__() or ''
            return columns, columns.__doc__ or ''
        # if it was a single string it was already made a list, and it's not a
        # callable, so it's either an iterable of labels... or
        if hasattr(columns, '__iter__'):
            return columns, ', '.join(str(elem) for elem in columns)
        # a single non-string label.
        return [columns], str(columns)

    def __init__(
            self, columns, exclude_columns=None, desc_temp=None,
            none_columns='error', **kwargs):
        self._exclude_columns = exclude_columns
        if exclude_columns:
            self._exclude_columns = self._interpret_columns_param(
                exclude_columns)
        self._none_error = False
        self._none_cols = None
        # handle none_columns
        if isinstance(none_columns, str):
            if none_columns == 'error':
                self._none_error = True
            elif none_columns == 'all':
                self._none_cols = AllColumns()
            else:
                raise ValueError((
                    "'error' and 'all' are the only valid string arguments"
                    " to the none_columns constructor parameter!"))
        elif hasattr(none_columns, '__iter__'):
            self._none_cols = none_columns
        elif callable(none_columns):
            self._none_cols = none_columns
        else:
            raise ValueError((
                "Valid arguments to the none_columns constructor parameter"
                " are 'error', 'all', an iterable of labels or a callable!"
            ))
        # done handling none_columns
        self._col_arg, self._col_str = self._interpret_columns_param(
            columns, self._none_error, none_columns=self._none_cols)
        if (kwargs.get('desc') is None) and desc_temp:
            kwargs['desc'] = desc_temp.format(self._col_str)
        if kwargs.get('exmsg') is None:
            kwargs['exmsg'] = (
                'Pipeline stage failed because not all columns {} '
                'were found in the input dataframe.'
            ).format(self._col_str)
        super().__init__(**kwargs)

    def _is_fittable(self):
        return is_fittable_column_qualifier(self._col_arg)

    @staticmethod
    def __get_cols_by_arg(col_arg, df, fit=False):
        try:
            if fit:
                # try to treat col_arg as a fittable column qualifier
                return col_arg.fit_transform(df)
            # else, no need to fit, so try to treat _col_arg as a callable
            return col_arg(df)
        except AttributeError:
            # got here cause col_arg has no fit_transform method...
            try:
                # so try and treat it as a callable again
                return col_arg(df)
            except TypeError:
                # calling col_arg 2 lines above failed; its a list of labels
                return col_arg
        except TypeError:
            # calling _col_arg 10 lines above failed; its a list of labels
            return col_arg

    def _get_columns(self, df, fit=False):
        cols = ColumnsBasedPipelineStage.__get_cols_by_arg(
            self._col_arg, df, fit=fit)
        if self._exclude_columns:
            exc_cols = ColumnsBasedPipelineStage.__get_cols_by_arg(
                self._exclude_columns, df, fit=fit)
            return [x for x in cols if x not in exc_cols]
        return cols

    def _prec(self, df):
        return set(self._get_columns(df=df)).issubset(df.columns)

    @abc.abstractmethod
    def _transformation(self, df, verbose, fit):
        raise NotImplementedError((
            "Classes extending ColumnsBasedPipelineStage must implement the "
            "_transformation method!"))

    def _fit_transform(self, df, verbose):
        self.is_fitted = True
        return self._transformation(df, verbose, fit=True)

    def _transform(self, df, verbose):
        return self._transformation(df, verbose, fit=False)


def _always_true(x):
    return True


class AdHocStage(PdPipelineStage):
    """An ad-hoc stage of a pandas DataFrame-processing pipeline.

    The signature for both the `transform` and the optional `fit_transform`
    callables is adaptive: The first argument is used positionally (so no
    specific name is assumed or used) to supply the callable with the pandas
    DataFrame object to transform. The following additional keyword arguments
    are supplied if the are included in the callable's signature:
    `verbose` - Passed on from pdpipe's `fit`, `fit_transform`
    and `apply` methods.
    `fit_context` and `application_context` - Provides fit-specific and
    application-specific contexts, in the form of PdpApplicationContext
    objects, usually available to pipeline stages using `self.fit_context` and
    `self.application_context`.

    Parameters
    ----------
    transform : callable
        The transformation this stage applies to dataframes. If the
        fit_transform parameter is also populated than this transformation is
        only applied on calls to transform. See documentation for the exact
        signature.
    fit_transform : callable, optional
        The transformation this stage applies to dataframes, only on
        fit_transform. Optional. See documentation for the exact signature.
    prec : callable, default None
        A callable that returns a boolean value. Represent a a precondition
        used to determine whether this stage can be applied to a given
        dataframe. If None is given, set to a function always returning True.
    """

    def __init__(self, transform, fit_transform=None, prec=None, **kwargs):
        if prec is None:
            prec = _always_true
        self._adhoc_transform = transform
        self._adhoc_fit_transform = fit_transform
        self._adhoc_prec = prec
        self._transform_kwargs = _get_args_list(self._adhoc_transform)
        try:
            self._fit_transform_kwargs = _get_args_list(
                self._adhoc_fit_transform)
        except TypeError:  # fit_transform is None
            self._fit_transform_kwargs = {}
        super().__init__(**kwargs)

    def _prec(self, df):
        return self._adhoc_prec(df)

    def _fit_transform(self, df, verbose):
        self.is_fitted = True
        if self._adhoc_fit_transform is None:
            self.is_fitted = True
            return self._transform(df, verbose)
        kwargs = {
            'verbose': verbose,
            'fit_context': self.fit_context,
            'application_context': self.application_context,
        }
        kwargs = {
            k: v for k, v in kwargs.items() if k in self._fit_transform_kwargs}
        return self._adhoc_fit_transform(df, **kwargs)

    def _transform(self, df, verbose):
        kwargs = {
            'verbose': verbose,
            'fit_context': self.fit_context,
            'application_context': self.application_context,
        }
        kwargs = {
            k: v for k, v in kwargs.items() if k in self._transform_kwargs}
        return self._adhoc_transform(df, **kwargs)


class PdPipeline(PdPipelineStage, collections.abc.Sequence):
    """A pipeline for processing pandas DataFrame objects.

    transformer_getter is usefull to avoid applying pipeline stages that are
    aimed to filter out items in a big dataset to create a training set for a
    machine learning model, for example, but should not be applied on future
    individual items to be transformed by the fitted pipeline.

    Parameters
    ----------
    stages : list
        A list of PdPipelineStage objects making up this pipeline.
    transform_getter : callable, optional
        A callable that can be applied to the fitted pipeline to produce a
        sub-pipeline of it which should be used to transform dataframes after
        the pipeline has been fitted. If not given, the fitted pipeline is used
        entirely.
    """

    _DEF_EXC_MSG = 'Pipeline precondition failed!'

    def __init__(self, stages, transformer_getter=None, **kwargs):
        self._stages = stages
        self._trans_getter = transformer_getter
        self.is_fitted = False
        super_kwargs = {
            'exraise': False,
            'exmsg': PdPipeline._DEF_EXC_MSG,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    # implementing a collections.abc.Sequence abstract method
    def __getitem__(self, index):
        if isinstance(index, slice):
            return PdPipeline(self._stages[index])

        if isinstance(index, list) and all(isinstance(x, str) for x in index):
            stages = [stage for stage in self._stages if stage._name in index]
            return PdPipeline(stages)

        if isinstance(index, str):
            stages = [stage for stage in self._stages if stage._name == index]
            if len(stages) == 0:
                raise ValueError(f"'{index}' is not exist.")
            return stages[0]

        return self._stages[index]

    # implementing a collections.abc.Sequence abstract method
    def __len__(self):
        return len(self._stages)

    def _prec(self, df):
        # PdPipeline overrides apply in a way which makes this moot
        raise NotImplementedError

    def _post(self, df):
        # PdPipeline overrides apply in a way which makes this moot
        raise NotImplementedError

    def _transform(self, df, verbose):
        # PdPipeline overrides apply in a way which makes this moot
        raise NotImplementedError

    def _post_transform_lock(self):
        self.application_context.lock()
        self.fit_context.lock()

    def apply(self, df, exraise=None, verbose=False):
        self.application_context = PdpApplicationContext()
        if self.is_fitted:
            res = self.transform(X=df, exraise=exraise, verbose=verbose)
            self._post_transform_lock()
            return res
        self.fit_context = PdpApplicationContext()
        res = self.fit_transform(X=df, exraise=exraise, verbose=verbose)
        self._post_transform_lock()
        return res

    def __timed_fit_transform(self, X, y=None, exraise=None, verbose=None):
        self.application_context = PdpApplicationContext()
        self.fit_context = PdpApplicationContext()
        inter_x = X
        times = []
        prev = time.time()
        for i, stage in enumerate(self._stages):
            try:
                stage.fit_context = self.fit_context
                stage.application_context = self.application_context
                inter_x = stage.fit_transform(
                    X=inter_x,
                    y=None,
                    exraise=exraise,
                    verbose=verbose,
                )
                now = time.time()
                times.append(now - prev)
                prev = now
            except Exception as e:
                raise PipelineApplicationError(
                    f"Exception raised in stage [ {i}] {stage}"
                ) from e
        self.is_fitted = True
        print("\nPipeline total application time: {:.3f}s.\n Details:".format(
            sum(times)))
        print(self.__times_str__(times))
        self._post_transform_lock()
        return inter_x

    def fit_transform(self, X, y=None, exraise=None, verbose=None, time=False):
        """Fits this pipeline and transforms the input dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to transform and fit this pipeline by.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Determines behaviour if the precondition of composing stages is not
            fulfilled by the input dataframe: If True, a
            pdpipe.FailedPreconditionError is raised. If False, the stage is
            skipped. If not given, or set to None, the default behaviour of
            each stage is used, as determined by its 'exraise' constructor
            parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            of each stage is checked but before its application. Otherwise, no
            messages are printed.
        time : bool, default False
            If True, per-stage application time is measured and reported when
            pipeline application is done.

        Returns
        -------
        pandas.DataFrame
            The resulting dacaframe.
        """
        if time:
            return self.__timed_fit_transform(
                X=X, y=y, exraise=exraise, verbose=verbose)
        inter_x = X
        self.application_context = PdpApplicationContext()
        self.fit_context = PdpApplicationContext()
        for i, stage in enumerate(self._stages):
            try:
                stage.fit_context = self.fit_context
                stage.application_context = self.application_context
                inter_x = stage.fit_transform(
                    X=inter_x,
                    y=None,
                    exraise=exraise,
                    verbose=verbose,
                )
            except Exception as e:
                raise PipelineApplicationError(
                    f"Exception raised in stage [ {i}] {stage}"
                ) from e
        self._post_transform_lock()
        self.is_fitted = True
        return inter_x

    def fit(self, X, y=None, exraise=None, verbose=None, time=None):
        """Fits this pipeline without transforming the input dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to fit this pipeline by.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Determines behaviour if the precondition of composing stages is not
            fulfilled by the input dataframe: If True, a
            pdpipe.FailedPreconditionError is raised. If False, the stage is
            skipped. If not given, or set to None, the default behaviour of
            each stage is used, as determined by its 'exraise' constructor
            parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            of each stage is checked but before its application. Otherwise, no
            messages are printed.
        time : bool, default False
            If True, per-stage application time is measured and reported when
            pipeline application is done.

        Returns
        -------
        pandas.DataFrame
            The input dataframe, unchanged.
        """
        self.fit_transform(
            X=X,
            y=None,
            exraise=exraise,
            verbose=verbose,
            time=time,
        )
        return X

    def __timed_transform(self, X, y=None, exraise=None, verbose=None):
        inter_x = X
        times = []
        prev = time.time()
        self.application_context = PdpApplicationContext()
        self.fit_context = PdpApplicationContext()
        for i, stage in enumerate(self._stages):
            try:
                stage.fit_context = self.fit_context
                stage.application_context = self.application_context
                inter_x = stage.transform(
                    X=inter_x,
                    y=None,
                    exraise=exraise,
                    verbose=verbose,
                )
                now = time.time()
                times.append(now - prev)
                prev = now
            except Exception as e:
                raise PipelineApplicationError(
                    f"Exception raised in stage [ {i}] {stage}"
                ) from e
        self.is_fitted = True
        print("\nPipeline total application time: {:.3f}s.\n Details:".format(
            sum(times)))
        print(self.__times_str__(times))
        self._post_transform_lock()
        return inter_x

    def transform(self, X, y=None, exraise=None, verbose=None, time=False):
        """Transforms the given dataframe without fitting this pipeline.

        If any stage in this pipeline is fittable but is not fitted, an
        UnfittedPipelineStageError is raised before transformation starts.

        Parameters
        ----------
        X : pandas.DataFrame
            The dataframe to transform.
        y : array-like, optional
            Targets for supervised learning.
        exraise : bool, default None
            Determines behaviour if the precondition of composing stages is not
            fulfilled by the input dataframe: If True, a
            pdpipe.FailedPreconditionError is raised. If False, the stage is
            skipped. If not given, or set to None, the default behaviour of
            each stage is used, as determined by its 'exraise' constructor
            parameter.
        verbose : bool, default False
            If True an explanation message is printed after the precondition
            of each stage is checked but before its application. Otherwise, no
            messages are printed.
        time : bool, default False
            If True, per-stage application time is measured and reported when
            pipeline application is done.

        Returns
        -------
        pandas.DataFrame
            The resulting dataframe.
        """
        for stage in self._stages:
            if stage._is_fittable() and not stage.is_fitted:
                raise UnfittedPipelineStageError((
                    "PipelineStage {} in pipeline is fittable but"
                    " unfitted!").format(stage))
        if time:
            return self.__timed_transform(
                X=X, y=y, exraise=exraise, verbose=verbose)
        inter_df = X
        self.application_context = PdpApplicationContext()
        for i, stage in enumerate(self._stages):
            try:
                stage.application_context = self.application_context
                inter_df = stage.transform(
                    X=inter_df,
                    y=None,
                    exraise=exraise,
                    verbose=verbose,
                )
            except Exception as e:
                raise PipelineApplicationError(
                    f"Exception raised in stage [ {i}] {stage}"
                ) from e
        self._post_transform_lock()
        return inter_df

    __call__ = apply

    def __add__(self, other):
        if isinstance(other, PdPipeline):
            return PdPipeline([*self._stages, *other._stages])
        if isinstance(other, PdPipelineStage):
            return PdPipeline([*self._stages, other])
        return NotImplemented

    def __times_str__(self, times):
        res = "A pdpipe pipeline:\n"
        stime = sum(times)
        if stime > 0:  # pragma: no cover
            percentages = [100 * x / stime for x in times]
        else:  # pragma: no cover
            percentages = [0 for x in times]
        res += '[ 0] [{:0>5.2f}s ({:0>5.2f}%)]  '.format(
            times[0], percentages[0]
        ) + "\n      ".join(
            textwrap.wrap(self._stages[0].description())
        ) + '\n'
        for i, stage in enumerate(self._stages[1:]):
            res += '[{:>2}] [{:0>5.2f}s ({:0>5.2f}%)]  '.format(
                i + 1, times[i + 1], percentages[i + 1]
            ) + "\n      ".join(
                textwrap.wrap(stage.description())
            ) + '\n'
        return res

    def __str__(self):
        res = "A pdpipe pipeline:\n"
        res += '[ 0]  ' + "\n      ".join(
            textwrap.wrap(self._stages[0].description())) + '\n'
        for i, stage in enumerate(self._stages[1:]):
            res += '[{:>2}]  '.format(i + 1) + "\n      ".join(
                textwrap.wrap(stage.description())) + '\n'
        return res

    def _mem_str(self, total):
        total = asizeof(self)
        lines = []
        for i, stage in enumerate(self._stages):
            size = asizeof(stage)
            if size > 500000:  # pragma: no cover
                lines.append('[{:>2}] {:.2f}Mb ({:0>5.2f}%), {}\n'.format(
                    i, size / 1000000, 100 * size / total,
                    stage.description()))
            elif size > 1000:  # pragma: no cover
                lines.append('[{:>2}] {:.2f}Kb ({:0>5.2f}%), {}\n'.format(
                    i, size / 1000, 100 * size / total, stage.description()))
            else:
                lines.append('[{:>2}] {:}b ({:0>5.2f}%), {}\n'.format(
                    i, size, 100 * size / total, stage.description()))
            lines.append(stage._mem_str())
        return ''.join(lines)

    def memory_report(self):
        """Prints a detailed memory report of the pipeline object to screen.

        To get better memory estimates make sure the pympler Python package is
        installed. Without it, sys.getsizeof is used, which can be extremely
        underestimate memory size of Python objects.
        """
        print("=== Pipeline memory report ===")
        size = asizeof(self)
        if size > 500000:  # pragma: no cover
            print("Total pipeline size in memory: {:.2f}Mb".format(
                size / 1000000))
        elif size > 1000:  # pragma: no cover
            print("Total pipeline size in memory: {:.2f}Kb".format(
                size / 1000))
        else:
            print("Total pipeline size in memory: {:.2f}b".format(
                size))
        print("Per-stage memory structure:")
        print(self._mem_str(total=size))

    def get_transformer(self):
        """Return the transformer induced by this fitted pipeline.

           This transformer is a `pdpipe` pipeline that transforms input data
           in a way corresponding to this pipline after it has been fitted. By
           default this is the pipeline itself, but the `transform_getter`
           constructor parameter can be used to return a sub-pipeline of the
           fitted pipeline instead, for cases where some stages should only be
           applied when fitting this pipeline to data.

        Returns
        -------
        pdpipe.PdPipeline
            The corresponding transformer pipeline induced by this pipeline.
        """
        try:
            return self._trans_getter(self)
        except TypeError:  # pragma: no cover
            return self

    # def drop(self, index):
    #     """Returns this pipeline with the stage of the given index removed.
    #     Arguments
    #     ---------
    #     index


def make_pdpipeline(*stages):
    """Constructs a PdPipeline from the given pipeline stages.

    Parameters
    ----------
    *stages : pdpipe.PipelineStage objects
       PdPipeline stages given as positional arguments.

    Returns
    -------
    p : pdpipe.PdPipeline
        The resulting pipeline.

    Examples
    --------
    import pdpipe as pdp
    make_pdpipeline(pdp.ColDrop('a'), pdp.Bin('speed'))
    """
    return PdPipeline(stages=stages)



"""Test the SnowballStem pipeline stage."""

import pytest
import pandas as pd
import pdpipe as pdp


@pytest.mark.first
def test_snowball_stem():
    df = pd.DataFrame([[3.2, ['kicking', 'boats']]], [1], ['freq', 'txt'])
    stem = pdp.SnowballStem('EnglishStemmer', 'txt')
    res_df = stem(df)
    assert 'txt' in res_df.columns
    assert 'txt_stem' not in res_df.columns
    assert res_df['txt'][1] == ['kick', 'boat']

    stem = pdp.SnowballStem('EnglishStemmer', 'txt', drop=False)
    res_df = stem(df)
    assert 'txt' in res_df.columns
    assert 'txt_stem' in res_df.columns
    assert res_df['txt'][1] == ['kicking', 'boats']
    assert res_df['txt_stem'][1] == ['kick', 'boat']


@pytest.mark.first
def test_snowball_stem_cond():
    df = pd.DataFrame([[3.2, ['kicking', 'boats']]], [1], ['freq', 'txt'])
    stem = pdp.SnowballStem('EnglishStemmer', 'txt', min_len=7)
    res_df = stem(df)
    assert 'txt' in res_df.columns
    assert 'txt_stem' not in res_df.columns
    assert res_df['txt'][1] == ['kick', 'boats']

    stem = pdp.SnowballStem('EnglishStemmer', 'txt', min_len=5)
    res_df = stem(df)
    assert 'txt' in res_df.columns
    assert 'txt_stem' not in res_df.columns
    assert res_df['txt'][1] == ['kick', 'boat']

    stem = pdp.SnowballStem('EnglishStemmer', 'txt', max_len=5)
    res_df = stem(df)
    assert 'txt' in res_df.columns
    assert 'txt_stem' not in res_df.columns
    assert res_df['txt'][1] == ['kicking', 'boat']

    df = pd.DataFrame(
        data=[[3.2, ['boats', 'kicking', 'squealing']]],
        index=[1],
        columns=['freq', 'txt'],
    )
    stem = pdp.SnowballStem('EnglishStemmer', 'txt', min_len=6, max_len=7)
    res_df = stem(df)
    assert 'txt' in res_df.columns
    assert 'txt_stem' not in res_df.columns
    assert res_df['txt'][1] == ['boats', 'kick', 'squealing']



"""Testing ApplyByCols pipeline stages."""

import math

import pytest
import pandas as pd

from pdpipe.col_generation import ApplyByCols


def ph_df():
    return pd.DataFrame(
        data=[[3.2, "acd"], [7.2, "alk"], [12.1, "alk"]],
        index=[1, 2, 3],
        columns=["ph", "lbl"]
    )


def test_applybycols():
    """Testing ApplyByCols pipeline stages."""
    df = ph_df()
    round_ph = ApplyByCols("ph", math.ceil)
    res_df = round_ph(df)
    assert res_df.columns.get_loc('ph') == 0
    assert res_df['ph'][1] == 4
    assert res_df['ph'][2] == 8
    assert res_df['ph'][3] == 13


def test_applybycols_func_desc():
    """Testing ApplyByCols pipeline stages."""
    df = ph_df()
    round_ph = ApplyByCols("ph", math.ceil, func_desc='Round PH values')
    res_df = round_ph(df)
    assert res_df.columns.get_loc('ph') == 0
    assert res_df['ph'][1] == 4
    assert res_df['ph'][2] == 8
    assert res_df['ph'][3] == 13


def test_applybycols_with_result_columns():
    """Testing ApplyByCols pipeline stages."""
    df = ph_df()
    round_ph = ApplyByCols("ph", math.ceil, result_columns='round_ph')
    res_df = round_ph(df)
    assert 'ph' not in res_df.columns
    assert res_df.columns.get_loc('round_ph') == 0
    assert res_df['round_ph'][1] == 4
    assert res_df['round_ph'][2] == 8
    assert res_df['round_ph'][3] == 13


def test_applybycols_with_drop():
    """Testing ApplyByCols pipeline stages."""
    df = ph_df()
    round_ph = ApplyByCols("ph", math.ceil, drop=False)
    res_df = round_ph(df)
    assert 'ph' in res_df.columns
    assert 'ph_app' in res_df.columns
    assert res_df.columns.get_loc('ph') == 0
    assert res_df.columns.get_loc('ph_app') == 1
    assert res_df['ph_app'][1] == 4
    assert res_df['ph_app'][2] == 8
    assert res_df['ph_app'][3] == 13


def test_applybycols_with_bad_len_result_columns():
    """Testing ApplyByCols pipeline stages."""
    with pytest.raises(ValueError):
        ApplyByCols("ph", math.ceil, result_columns=['a', 'b'])


# def _num_df():
#     return pd.DataFrame(
#         data=[[1, 2, 'a'], [2, 4, 'b']],
#         index=[1, 2],
#         columns=['num1', 'num2', 'char']
#     )


# def _sum_and_diff(row):
#     return pd.Series({
#         'sum': row['num1'] + row['num2'],
#         'diff': row['num1'] - row['num2']
#     })


# def test_applytorows_with_df_generation():
#     """Testing ApplyToRows pipeline stages."""
#     df = _num_df()
#     cbf_stage = ApplyToRows(_sum_and_diff)
#     res_df = cbf_stage(df)
#     assert 'sum' in res_df.columns
#     assert 'diff' in res_df.columns
#     assert res_df['sum'][1] == 3
#     assert res_df['sum'][2] == 6
#     assert res_df['diff'][1] == -1
#     assert res_df['diff'][2] == -2


# def test_applytorows_with_df_generation_and_optionals():
#     """Testing ApplyToRows pipeline stages."""
#     df = _num_df()
#     cbf_stage = ApplyToRows(
#         func=_sum_and_diff,
#         func_desc="calculates sum and diff of num1 and num2",
#         prec=lambda df: 'num1' in df.columns and 'num2' in df.columns
#     )
#     res_df = cbf_stage(df)
#     assert 'sum' in res_df.columns
#     assert 'diff' in res_df.columns
#     assert res_df['sum'][1] == 3
#     assert res_df['sum'][2] == 6
#     assert res_df['diff'][1] == -1
#     assert res_df['diff'][2] == -2


# def test_applytorows_with_df_generation_follow():
#     """Testing ApplyToRows pipeline stages."""
#     df = _num_df()
#     cbf_stage = ApplyToRows(_sum_and_diff, follow_column='num1')
#     res_df = cbf_stage(df)
#     assert res_df.columns.get_loc('sum') == 2
#     assert res_df.columns.get_loc('diff') == 1
#     assert 'sum' in res_df.columns
#     assert 'diff' in res_df.columns
#     assert res_df['sum'][1] == 3
#     assert res_df['sum'][2] == 6
#     assert res_df['diff'][1] == -1
#     assert res_df['diff'][2] == -2



"""Testing AggByCols pipeline stages."""

import pytest
import numpy as np
from numpy.testing import assert_approx_equal
import pandas as pd

from pdpipe import AggByCols


_LOG32 = 1.163151
_LOG72 = 1.974081
_LOG121 = 2.493205


def ph_df():
    return pd.DataFrame(
        data=[[3.2, "acd"], [7.2, "alk"], [12.1, "alk"]],
        index=[1, 2, 3],
        columns=["ph", "lbl"]
    )


def test_aggbycols():
    """Testing AggByCols pipeline stages."""
    df = ph_df()
    round_ph = AggByCols("ph", np.log)
    res_df = round_ph(df)
    assert res_df.columns.get_loc('ph') == 0
    assert_approx_equal(res_df['ph'][1], _LOG32, significant=5)
    assert_approx_equal(res_df['ph'][2], _LOG72, significant=5)
    assert_approx_equal(res_df['ph'][3], _LOG121, significant=5)


def test_aggbycols_func_desc():
    """Testing AggByCols pipeline stages."""
    df = ph_df()
    round_ph = AggByCols("ph", np.log, func_desc='Round PH values')
    res_df = round_ph(df)
    assert res_df.columns.get_loc('ph') == 0
    assert_approx_equal(res_df['ph'][1], _LOG32, significant=5)
    assert_approx_equal(res_df['ph'][2], _LOG72, significant=5)
    assert_approx_equal(res_df['ph'][3], _LOG121, significant=5)


def test_aggbycols_with_result_columns():
    """Testing AggByCols pipeline stages."""
    df = ph_df()
    round_ph = AggByCols("ph", np.log, result_columns='log_ph')
    res_df = round_ph(df)
    assert 'ph' not in res_df.columns
    assert res_df.columns.get_loc('log_ph') == 0
    assert_approx_equal(res_df['log_ph'][1], _LOG32, significant=5)
    assert_approx_equal(res_df['log_ph'][2], _LOG72, significant=5)
    assert_approx_equal(res_df['log_ph'][3], _LOG121, significant=5)


def test_aggbycols_with_drop():
    """Testing AggByCols pipeline stages."""
    df = ph_df()
    round_ph = AggByCols("ph", np.log, drop=False)
    res_df = round_ph(df)
    assert 'ph' in res_df.columns
    assert 'ph_agg' in res_df.columns
    assert res_df.columns.get_loc('ph') == 0
    assert res_df.columns.get_loc('ph_agg') == 1
    assert_approx_equal(res_df['ph_agg'][1], _LOG32, significant=5)
    assert_approx_equal(res_df['ph_agg'][2], _LOG72, significant=5)
    assert_approx_equal(res_df['ph_agg'][3], _LOG121, significant=5)


def test_aggbycols_no_drop_custom_suffix():
    """Testing AggByCols pipeline stages."""
    df = ph_df()
    round_ph = AggByCols("ph", np.log, drop=False, suffix='_log')
    res_df = round_ph(df)
    assert 'ph' in res_df.columns
    assert 'ph_log' in res_df.columns
    assert res_df.columns.get_loc('ph') == 0
    assert res_df.columns.get_loc('ph_log') == 1
    assert_approx_equal(res_df['ph_log'][1], _LOG32, significant=5)
    assert_approx_equal(res_df['ph_log'][2], _LOG72, significant=5)
    assert_approx_equal(res_df['ph_log'][3], _LOG121, significant=5)


def test_aggbycols_with_bad_len_result_columns():
    """Testing ApplyByCols pipeline stages."""
    with pytest.raises(ValueError):
        AggByCols("ph", np.log, result_columns=['a', 'b'])



"""Testing basic pipeline stages."""

import pytest
import pandas as pd

import pdpipe as pdp
from pdpipe.sklearn_stages import Encode
from pdpipe.exceptions import UnfittedPipelineStageError


def _some_df():
    return pd.DataFrame(
        data=[[3.2, "acd", "x1"], [7.2, "alk", "x2"], [12.1, "alk", "x3"]],
        index=[1, 2, 3],
        columns=["ph", "lbl", "name"]
    )


def _some_df2():
    return pd.DataFrame(
        data=[[4.4, "alk", "x1"], [6.1, "acd", "x2"], [1.2, "alk", "x3"]],
        index=[1, 2, 3],
        columns=["ph", "lbl", "name"]
    )


def test_encode():
    df = _some_df()
    encode_stage = Encode()
    res_df = encode_stage(df)
    assert 'lbl' in res_df.columns
    assert res_df['lbl'][1] == 0
    assert res_df['lbl'][2] == 1
    assert res_df['lbl'][3] == 1
    assert res_df['name'][1] == 0
    assert res_df['name'][2] == 1
    assert res_df['name'][3] == 2

    # see only transform (no fit) when already fitted
    df2 = _some_df2()
    res_df2 = encode_stage(df2)
    assert 'lbl' in res_df2.columns
    assert res_df2['lbl'][1] == 1
    assert res_df2['lbl'][2] == 0
    assert res_df2['lbl'][3] == 1
    assert res_df2['name'][1] == 0
    assert res_df2['name'][2] == 1
    assert res_df2['name'][3] == 2

    # see only transform (no fit) when already fitted
    df2 = _some_df2()
    res_df2 = encode_stage.transform(df2)
    assert 'lbl' in res_df2.columns
    assert res_df2['lbl'][1] == 1
    assert res_df2['lbl'][2] == 0
    assert res_df2['lbl'][3] == 1
    assert res_df2['name'][1] == 0
    assert res_df2['name'][2] == 1
    assert res_df2['name'][3] == 2

    # check fit_transform when already fitted
    df2 = _some_df2()
    res_df2 = encode_stage.fit_transform(df2)
    assert 'lbl' in res_df.columns
    assert res_df2['lbl'][1] == 1
    assert res_df2['lbl'][2] == 0
    assert res_df2['lbl'][3] == 1
    assert res_df2['name'][1] == 0
    assert res_df2['name'][2] == 1
    assert res_df2['name'][3] == 2


def test_encode_fit():
    df = _some_df()
    encode_stage = Encode()

    with pytest.raises(UnfittedPipelineStageError):
        encode_stage.transform(df)

    res_df = encode_stage.fit(df)
    assert 'lbl' in res_df.columns
    assert res_df['lbl'][1] == 'acd'
    assert res_df['lbl'][2] == 'alk'
    assert res_df['lbl'][3] == 'alk'

    # see only transform (no fit) when already fitted
    df2 = _some_df2()
    res_df2 = encode_stage.transform(df2)
    assert 'lbl' in res_df2.columns
    assert res_df2['lbl'][1] == 1
    assert res_df2['lbl'][2] == 0
    assert res_df2['lbl'][3] == 1
    assert res_df2['name'][1] == 0
    assert res_df2['name'][2] == 1
    assert res_df2['name'][3] == 2

    # check that apply only transforms (no fit) when already fitted
    df2 = _some_df2()
    res_df2 = encode_stage.transform(df2)
    assert 'lbl' in res_df2.columns
    assert res_df2['lbl'][1] == 1
    assert res_df2['lbl'][2] == 0
    assert res_df2['lbl'][3] == 1
    assert res_df2['name'][1] == 0
    assert res_df2['name'][2] == 1
    assert res_df2['name'][3] == 2


def test_encode_with_args():
    df = _some_df()
    encode_stage = Encode("lbl", drop=False)
    res_df = encode_stage(df, verbose=True)
    assert 'lbl' in res_df.columns
    assert res_df['lbl_enc'][1] == 0
    assert res_df['lbl_enc'][2] == 1
    assert res_df['lbl_enc'][3] == 1

    # see only transform (no fit) when already fitted
    df2 = _some_df2()
    res_df2 = encode_stage(df2)
    assert 'lbl' in res_df.columns
    assert res_df2['lbl_enc'][1] == 1
    assert res_df2['lbl_enc'][2] == 0
    assert res_df2['lbl_enc'][3] == 1

    # check fit_transform when already fitted
    df2 = _some_df2()
    res_df2 = encode_stage.fit_transform(df2, verbose=True)
    assert 'lbl' in res_df.columns
    assert res_df2['lbl_enc'][1] == 1
    assert res_df2['lbl_enc'][2] == 0
    assert res_df2['lbl_enc'][3] == 1


def test_encode_with_exclude():
    df = _some_df()
    encode_stage = Encode("lbl", exclude_columns="name")
    res_df = encode_stage(df)
    assert 'lbl' in res_df.columns
    assert res_df['lbl'][1] == 0
    assert res_df['lbl'][2] == 1
    assert res_df['lbl'][3] == 1


def test_encode_in_pipeline():
    drop_name = pdp.ColDrop('name')
    encode_stage = Encode()
    pline = drop_name + encode_stage

    df = _some_df()
    res_df = pline(df)
    assert 'lbl' in res_df.columns
    assert 'name' not in res_df.columns
    assert res_df['lbl'][1] == 0
    assert res_df['lbl'][2] == 1
    assert res_df['lbl'][3] == 1

    # check fitted pipeline
    df2 = _some_df2()
    res_df2 = pline(df2)
    assert 'lbl' in res_df2.columns
    assert res_df2['lbl'][1] == 1
    assert res_df2['lbl'][2] == 0
    assert res_df2['lbl'][3] == 1


def test_encode_in_pipelin_fit_n_transform():
    drop_name = pdp.ColDrop('name')
    encode_stage = Encode()
    pline = drop_name + encode_stage

    df = _some_df()

    with pytest.raises(UnfittedPipelineStageError):
        res_df = pline.transform(df)

    res_df = pline.fit(df)
    assert 'lbl' in res_df.columns
    assert 'name' in res_df.columns
    assert res_df['lbl'][1] == 'acd'
    assert res_df['lbl'][2] == 'alk'
    assert res_df['lbl'][3] == 'alk'

    res_df = pline.transform(df)
    assert 'lbl' in res_df.columns
    assert 'name' not in res_df.columns
    assert res_df['lbl'][1] == 0
    assert res_df['lbl'][2] == 1
    assert res_df['lbl'][3] == 1

    # check fitted pipeline
    df2 = _some_df2()
    res_df2 = pline.transform(df2)
    assert 'lbl' in res_df2.columns
    assert res_df2['lbl'][1] == 1
    assert res_df2['lbl'][2] == 0
    assert res_df2['lbl'][3] == 1

    # check fit_transform when already fitted
    df2 = _some_df2()
    res_df2 = pline.fit_transform(df2, verbose=True)
    assert 'lbl' in res_df.columns
    assert 'name' not in res_df.columns
    assert res_df2['lbl'][1] == 1
    assert res_df2['lbl'][2] == 0
    assert res_df2['lbl'][3] == 1



"""Testing attribute pipeline stages."""

import pandas as pd

import pdpipe as pdp
from pdpipe.core import PdPipeline
from pdpipe.basic_stages import ColDrop
from pdpipe.col_generation import Bin


def _some_df():
    return pd.DataFrame(
        data=[[-3, 'Will'], [4, 'Jan'], [5, 'Tasha'], [9, 'Data']],
        index=[1, 2, 3, 4],
        columns=['speed', 'name'])


def test_attribute_stage():
    """Testing attribute pipeline stages."""
    pipeline = pdp.ColDrop('name').Bin({'speed': [5]}, drop=True)
    assert isinstance(pipeline, PdPipeline)
    assert isinstance(pipeline[0], ColDrop)
    assert isinstance(pipeline[1], Bin)
    df = _some_df()
    res_df = pipeline(df)
    assert 'speed' in res_df.columns
    assert 'name' not in res_df.columns



"""PdPipeline stages dependent on the scikit-learn Python library.

Please note that the scikit-learn Python package must be installed for the
stages in this module to work.

When attempting to load stages from this module, pdpipe will first attempt to
import sklearn. If it fails, it will issue a warning, will not import any of
the pipeline stages that make up this module, and continue to load other
pipeline stages.
"""

import numpy as np
import pandas as pd
import sklearn.preprocessing
import tqdm
from skutil.preprocessing import scaler_by_params
from sklearn.feature_extraction.text import (
    TfidfVectorizer,
)

from pdpipe.core import PdPipelineStage, ColumnsBasedPipelineStage
from pdpipe.util import (
    out_of_place_col_insert,
    per_column_valus_sklearn_transform,
)
from pdpipe.cq import OfDtypes
from pdpipe.shared import (
    _get_args_list,
    _identity_function,
)

from .exceptions import PipelineApplicationError


class Encode(ColumnsBasedPipelineStage):
    """A pipeline stage that encodes categorical columns to integer values.

    The encoder for each column is saved in the attribute 'encoders', which
    is a dict mapping each encoded column name to the
    sklearn.preprocessing.LabelEncoder object used to encode it.

    Parameters
    ----------
    columns : single label, list-like or callable, default None
        Column labels in the DataFrame to be encoded. If columns is None then
        all the columns with object or category dtype will be converted, except
        those given in the exclude_columns parameter. Alternatively,
        this parameter can be assigned a callable returning an iterable of
        labels from an input pandas.DataFrame. See pdpipe.cq.
    exclude_columns : str or list-like, default None
        Label or labels of columns to be excluded from encoding. If None then
        no column is excluded. Alternatively, this parameter can be assigned a
        callable returning an iterable of labels from an input
    drop : bool, default True
        If set to True, the source columns are dropped after being encoded,
        and the resulting encoded columns retain the names of the source
        columns. Otherwise, encoded columns gain the suffix '_enc'.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> data = [[3.2, "acd"], [7.2, "alk"], [12.1, "alk"]]
        >>> df = pd.DataFrame(data, [1,2,3], ["ph","lbl"])
        >>> encode_stage = pdp.Encode("lbl")
        >>> encode_stage(df)
             ph  lbl
        1   3.2    0
        2   7.2    1
        3  12.1    1
        >>> encode_stage.encoders["lbl"].inverse_transform([0,1,1])
        array(['acd', 'alk', 'alk'], dtype=object)
    """

    def __init__(
        self, columns=None, exclude_columns=None, drop=True, **kwargs
    ):
        self._drop = drop
        self.encoders = {}
        super_kwargs = {
            'columns': columns,
            'exclude_columns': exclude_columns,
            'desc_temp': "Encode {}",
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = OfDtypes(['object', 'category'])
        super().__init__(**super_kwargs)

    def _transformation(self, df, verbose, fit):
        raise NotImplementedError

    def _fit_transform(self, df, verbose):
        self.encoders = {}
        columns_to_encode = self._get_columns(df, fit=True)
        if verbose:
            columns_to_encode = tqdm.tqdm(columns_to_encode)
        inter_df = df
        for colname in columns_to_encode:
            lbl_enc = sklearn.preprocessing.LabelEncoder()
            source_col = df[colname]
            loc = df.columns.get_loc(colname) + 1
            new_name = colname + "_enc"
            if self._drop:
                inter_df = inter_df.drop(colname, axis=1)
                new_name = colname
                loc -= 1
            inter_df = out_of_place_col_insert(
                df=inter_df,
                series=lbl_enc.fit_transform(source_col),
                loc=loc,
                column_name=new_name,
            )
            self.encoders[colname] = lbl_enc
        self.is_fitted = True
        return inter_df

    def _transform(self, df, verbose):
        inter_df = df
        for colname in self.encoders:
            lbl_enc = self.encoders[colname]
            source_col = df[colname]
            loc = df.columns.get_loc(colname) + 1
            new_name = colname + "_enc"
            if self._drop:
                inter_df = inter_df.drop(colname, axis=1)
                new_name = colname
                loc -= 1
            inter_df = out_of_place_col_insert(
                df=inter_df,
                series=lbl_enc.transform(source_col),
                loc=loc,
                column_name=new_name,
            )
        return inter_df


class Scale(ColumnsBasedPipelineStage):
    """A pipeline stage that scales data.

    Parameters
    ----------
    scaler : str
        The type of scaler to use to scale the data. One of 'StandardScaler',
        'MinMaxScaler', 'MaxAbsScaler', 'RobustScaler', 'QuantileTransformer'
        and 'Normalizer'.
    columns : single label, list-like or callable, default None
        Column labels in the DataFrame to be scale. If columns is None then
        all columns of numeric dtype will be scaled, except those given in the
        exclude_columns parameter. Alternatively, this parameter can be
        assigned a callable returning an iterable of labels from an input
        pandas.DataFrame. See pdpipe.cq.
    exclude_columns : str or list-like, optional
        Label or labels of columns to be excluded from encoding. Alternatively,
        this parameter can be assigned a callable returning an iterable of
        labels from an input pandas.DataFrame. See pdpipe.cq.
    joint : bool, default False
        If set to True, all scaled columns will be scaled as a single value
        set (meaning, only the single largest value among all input columns
        will be scaled to 1, and not the largest one for each column).
    **kwargs : extra keyword arguments
        All valid extra keyword arguments are forwarded to the scaler
        constructor on scaler creation (e.g. 'n_quantiles' for
        QuantileTransformer). PdPipelineStage valid keyword arguments are used
        to override Scale class defaults.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> data = [[3.2, 0.3], [7.2, 0.35], [12.1, 0.29]]
        >>> df = pd.DataFrame(data, [1,2,3], ["ph","gt"])
        >>> scale_stage = pdp.Scale("StandardScaler")
        >>> scale_stage(df)
                 ph        gt
        1 -1.181449 -0.508001
        2 -0.082427  1.397001
        3  1.263876 -0.889001
    """

    def __init__(
        self,
        scaler,
        columns=None,
        exclude_columns=None,
        joint=False,
        **kwargs
    ):
        self.scaler = scaler
        self.joint = joint
        self._kwargs = kwargs.copy()
        super_kwargs = {
            'columns': columns,
            'exclude_columns': exclude_columns,
            'desc_temp': "Scale columns {}",
        }
        valid_super_kwargs = super()._init_kwargs()
        for key in kwargs:
            if key in valid_super_kwargs:
                super_kwargs[key] = kwargs[key]
                self._kwargs.pop(key)
        super_kwargs['none_columns'] = OfDtypes([np.number])
        super().__init__(**super_kwargs)

    def _transformation(self, df, verbose, fit):
        raise NotImplementedError

    def _fit_transform(self, df, verbose):
        self._columns_to_scale = self._get_columns(df, fit=True)
        unscaled_cols = [
            x for x in df.columns
            if x not in self._columns_to_scale
        ]
        col_order = list(df.columns)
        inter_df = df[self._columns_to_scale]
        self._scaler = scaler_by_params(self.scaler, **self._kwargs)
        try:
            if self.joint:
                self._scaler.fit(np.array([inter_df.values.flatten()]).T)
                inter_df = per_column_valus_sklearn_transform(
                    df=inter_df,
                    transform=self._scaler.transform
                )
            else:
                inter_df = pd.DataFrame(
                    data=self._scaler.fit_transform(inter_df.values),
                    index=inter_df.index,
                    columns=inter_df.columns,
                )
        except Exception as e:
            raise PipelineApplicationError(
                "Exception raised when Scale applied to columns"
                f" {self._columns_to_scale} by class {self.__class__}"
            ) from e
        if len(unscaled_cols) > 0:
            unscaled = df[unscaled_cols]
            inter_df = pd.concat([inter_df, unscaled], axis=1)
            inter_df = inter_df[col_order]
        self.is_fitted = True
        return inter_df

    def _transform(self, df, verbose):
        unscaled_cols = [
            x for x in df.columns
            if x not in self._columns_to_scale
        ]
        col_order = list(df.columns)
        inter_df = df[self._columns_to_scale]
        try:
            if self.joint:
                inter_df = per_column_valus_sklearn_transform(
                    df=inter_df,
                    transform=self._scaler.transform
                )
            else:
                inter_df = pd.DataFrame(
                    data=self._scaler.transform(inter_df.values),
                    index=inter_df.index,
                    columns=inter_df.columns,
                )
        except Exception:
            raise PipelineApplicationError(
                "Exception raised when Scale applied to columns"
                f" {self._columns_to_scale} by class {self.__class__}"
            )
        if len(unscaled_cols) > 0:
            unscaled = df[unscaled_cols]
            inter_df = pd.concat([inter_df, unscaled], axis=1)
            inter_df = inter_df[col_order]
        return inter_df


class TfidfVectorizeTokenLists(PdPipelineStage):
    """A pipeline stage TFIDF-vectorizing a token-list column to count columns.

    Every cell in the input columns is assumed to be a list of strings, each
    representing a single token. The resulting TF-IDF vector is exploded into
    individual columns, each with the label 'lbl_i' where lbl is the original
    column label and i is the index of column in the count vector.

    The resulting columns are concatenated to the end of the dataframe.

    All valid sklearn.TfidfVectorizer keyword arguemnts can be provided as
    keyword arguments to the constructor, except 'input' and 'analyzer', which
    will be ignored. As usual, all valid PdPipelineStage constructor parameters
    can also be provided as keyword arguments.

    Parameters
    ----------
    column : str
        The label of the token-list column to TfIdf-vectorize.
    drop : bool, default True
        If set to True, the source column is dropped after being transformed.
    hierarchical_labels : bool, default False
        If set to True, the labels of resulting columns are of the form 'P_F'
        where P is the label of the original token-list column and F is the
        feature name (i.e. the string token it corresponds to). Otherwise, it
        is simply the feature name itself. If you plan to have two different
        TfidfVectorizeTokenLists pipeline stages vectorizing two different
        token-list columns, you should set this to true, so tf-idf features
        originating in different text columns do not overwrite one another.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> data = [[2, ['hovercraft', 'eels']], [5, ['eels', 'urethra']]]
        >>> df = pd.DataFrame(data, [1, 2], ['Age', 'tokens'])
        >>> tfvectorizer = pdp.TfidfVectorizeTokenLists('tokens')
        >>> tfvectorizer(df)
           Age      eels  hovercraft   urethra
        1    2  0.579739    0.814802  0.000000
        2    5  0.579739    0.000000  0.814802
    """

    _DEF_CNTVEC_MSG = "Count-vectorizing column {}."

    def __init__(self, column, drop=True, hierarchical_labels=False,
                 **kwargs):
        self._column = column
        self._drop = drop
        self._hierarchical_labels = hierarchical_labels
        msg = TfidfVectorizeTokenLists._DEF_CNTVEC_MSG.format(column)
        super_kwargs = {
            "exmsg": ("TfIdfVectorizeTokenLists precondition not met:"
                      f"{column} column not found."),
            "desc": msg,
        }
        valid_vectorizer_args = _get_args_list(TfidfVectorizer.__init__)
        self._vectorizer_args = {
            k: kwargs[k] for k in kwargs
            if k in valid_vectorizer_args and k not in [
                'input', 'analyzer', 'self',
            ]
        }
        pipeline_stage_args = {
            k: kwargs[k] for k in kwargs
            if k in PdPipelineStage._INIT_KWARGS
        }
        super_kwargs.update(**pipeline_stage_args)
        super().__init__(**super_kwargs)

    def _prec(self, df):
        return self._column in df.columns

    def _fit_transform(self, df, verbose):
        self._tfidf_vectorizer = TfidfVectorizer(
            input='content',
            analyzer=_identity_function,
            **self._vectorizer_args,
        )
        vectorized = self._tfidf_vectorizer.fit_transform(df[self._column])
        self._n_features = vectorized.shape[1]
        if self._hierarchical_labels:
            self._res_col_names = [
                f'{self._column}_{f}'
                for f in self._tfidf_vectorizer.get_feature_names()
            ]
        else:
            self._res_col_names = self._tfidf_vectorizer.get_feature_names()
        vec_df = pd.DataFrame.sparse.from_spmatrix(
            data=vectorized, index=df.index, columns=self._res_col_names)
        inter_df = pd.concat([df, vec_df], axis=1)
        self.is_fitted = True
        if self._drop:
            return inter_df.drop(self._column, axis=1)
        return inter_df

    def _transform(self, df, verbose):
        vectorized = self._tfidf_vectorizer.transform(df[self._column])
        vec_df = pd.DataFrame.sparse.from_spmatrix(
            data=vectorized, index=df.index, columns=self._res_col_names)
        inter_df = pd.concat([df, vec_df], axis=1)
        if self._drop:
            return inter_df.drop(self._column, axis=1)
        return inter_df



"""Testing basic pipeline stages."""

import pandas as pd

from pdpipe.basic_stages import DropNa


DF1 = pd.DataFrame([[1, 4], [4, None], [1, 11]], [1, 2, 3], ['a', 'b'])


def test_dropna_basic():
    """Testing the DropNa pipeline stage."""
    res_df = DropNa().apply(DF1)
    assert 1 in res_df.index
    assert 2 not in res_df.index
    assert 3 in res_df.index


def test_dropna_verbose():
    """Testing the DropNa pipeline stage."""
    res_df = DropNa().apply(DF1, verbose=True)
    assert 1 in res_df.index
    assert 2 not in res_df.index
    assert 3 in res_df.index



"""Testing basic pipeline stages."""

import pandas as pd
import pytest

import pdpipe as pdp
from pdpipe import ColDrop
from pdpipe.exceptions import FailedPreconditionError


def _test_df():
    return pd.DataFrame(
        data=[[1, 2, 'a'], [2, 4, 'b']],
        index=[1, 2],
        columns=['num1', 'num2', 'char']
    )


def test_coldrop_one_col():
    """Testing the ColDrop pipeline stage."""
    df = _test_df()
    assert 'num1' in df.columns
    stage = ColDrop('num1')
    res_df = stage.apply(df)
    assert 'num1' not in res_df.columns
    assert 'num2' in res_df.columns
    assert 'char' in res_df.columns

    # make sure fit is null operation for unfittable stages
    res_df = stage.fit(df)
    assert 'num1' in res_df.columns
    assert 'num2' in res_df.columns
    assert 'char' in res_df.columns

    res_df = stage.fit(df, verbose=True)
    assert 'num1' in res_df.columns
    assert 'num2' in res_df.columns
    assert 'char' in res_df.columns

    # make sure transform and fit_transform are equivalent to apply
    # for unfittable stages
    res_df = stage.transform(df)
    assert 'num1' not in res_df.columns
    assert 'num2' in res_df.columns
    assert 'char' in res_df.columns

    res_df = stage.transform(df, verbose=True)
    assert 'num1' not in res_df.columns
    assert 'num2' in res_df.columns
    assert 'char' in res_df.columns

    res_df = stage.fit_transform(df)
    assert 'num1' not in res_df.columns
    assert 'num2' in res_df.columns
    assert 'char' in res_df.columns

    res_df = stage.fit_transform(df, verbose=True)
    assert 'num1' not in res_df.columns
    assert 'num2' in res_df.columns
    assert 'char' in res_df.columns


def test_coldrop_missing_col():
    """Testing the ColDrop pipeline stage."""
    df = _test_df()
    assert 'num1' in df.columns
    stage = ColDrop('num3')
    with pytest.raises(FailedPreconditionError):
        res_df = stage.apply(df)

    res_df = stage.apply(df, exraise=False)
    assert res_df.equals(df)

    # make sure fit is null operation for unfittable stages
    with pytest.raises(FailedPreconditionError):
        res_df = stage.fit(df)

    with pytest.raises(FailedPreconditionError):
        res_df = stage.fit(df, verbose=True)

    res_df = stage.fit(df, exraise=False)
    assert res_df.equals(df)

    # make sure transform and fit_transform are equivalent to apply
    # for unfittable stages
    with pytest.raises(FailedPreconditionError):
        res_df = stage.transform(df)

    with pytest.raises(FailedPreconditionError):
        res_df = stage.transform(df, verbose=True)

    res_df = stage.transform(df, exraise=False)
    assert res_df.equals(df)

    with pytest.raises(FailedPreconditionError):
        res_df = stage.fit_transform(df)

    with pytest.raises(FailedPreconditionError):
        res_df = stage.fit_transform(df, verbose=True)

    res_df = stage.fit_transform(df, exraise=False)
    assert res_df.equals(df)

    stage = ColDrop('num3', errors='ignore')
    res_df = stage.apply(df)
    assert res_df.equals(df)

    # make sure fit is null operation for unfittable stages
    res_df = stage.fit(df)
    assert res_df.equals(df)

    # make sure transform and fit_transform are equivalent to apply
    # for unfittable stages
    res_df = stage.fit_transform(df)
    assert res_df.equals(df)

    res_df = stage.transform(df)
    assert res_df.equals(df)


def test_coldrop_multi_col():
    """Testing the ColDrop pipeline stage."""
    df = _test_df()
    assert 'num1' in df.columns
    assert 'num2' in df.columns
    stage = ColDrop(['num1', 'num2'])
    res_df = stage.apply(df)
    assert 'num1' not in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns


def test_coldrop_col_qualifier():
    """Testing the ColDrop pipeline stage."""
    df = _test_df()
    assert 'num1' in df.columns
    assert 'num2' in df.columns
    stage = ColDrop(pdp.cq.StartWith('num'))
    res_df = stage.apply(df)
    assert 'num1' not in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns


def _test_df2():
    return pd.DataFrame(
        data=[[1, 2, 'a'], [2, 4, 'b']],
        index=[1, 2],
        columns=['num1', 2, False]
    )


def test_coldrop_non_str_lbl():
    """Testing the ColDrop pipeline stage."""
    df = _test_df2()
    assert 2 in df.columns
    stage = ColDrop(2)
    res_df = stage.apply(df)
    assert 2 not in res_df.columns
    assert 'num1' in res_df.columns
    assert False in res_df.columns



"""Shared inner functionalities for pdpipe."""

import inspect


def _interpret_columns_param(columns):
    if isinstance(columns, str):
        return [columns]
    if hasattr(columns, '__iter__'):
        return columns
    return [columns]


def _list_str(listi):
    if listi is None:
        return None
    if isinstance(listi, (list, tuple)):
        return ', '.join([str(elem) for elem in listi])
    return listi


def _get_args_list(func):
    signature = inspect.signature(func)
    return list(signature.parameters.keys())


def _identity_function(x):
    return x
